#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Var,
    check_optimal_termination,
    Param,
    Suffix,
    NonNegativeReals,
    Reference,
    units as pyunits,
)
from pyomo.common.config import Bool, ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
    MaterialFlowBasis,
)
from idaes.core.solvers import get_solver
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError, InitializationError
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from watertap.core import ControlVolume0DBlock, InitializationMixin
from watertap.costing.unit_models.nanofiltration import cost_nanofiltration


_log = idaeslog.getLogger(__name__)


@declare_process_block_class("NanofiltrationZO")
class NanofiltrationData(InitializationMixin, UnitModelBlockData):
    """
    Zero order nanofiltration model based on specified water flux and ion rejection.
    Default data from Table 9 in Labban et al. (2017) https://doi.org/10.1016/j.memsci.2016.08.062
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. NF units do not support dynamic
    behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. NF units do not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "is_isothermal",
        ConfigValue(
            default=True,
            domain=Bool,
            description="""Assume isothermal conditions for control volume(s); energy_balance_type must be EnergyBalanceType.none,
    **default** - True.""",
        ),
    )

    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.none,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
    **default** - EnergyBalanceType.none.
    **Valid values:** {
    **EnergyBalanceType.useDefault - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material,
    **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
    **MomentumBalanceType.momentumTotal** - single momentum balance for material,
    **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
    constructed,
    **default** - False.
    **Valid values:** {
    **True** - include pressure change terms,
    **False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    def _process_config(self):
        if len(self.config.property_package.solvent_set) > 1:
            raise ConfigurationError(
                "NF model only supports one solvent component,"
                "the provided property package has specified {} solvent components".format(
                    len(self.config.property_package.solvent_set)
                )
            )

        if len(self.config.property_package.solvent_set) == 0:
            raise ConfigurationError(
                "The NF model was expecting a solvent and did not receive it."
            )

        if (
            len(self.config.property_package.solute_set) == 0
            and len(self.config.property_package.ion_set) == 0
        ):
            raise ConfigurationError(
                "The NF model was expecting at least one solute or ion and did not receive any."
            )

    def _validate_config(self):
        if (
            self.config.is_isothermal
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            raise ConfigurationError(
                "If the isothermal assumption is used then the energy balance type must be none"
            )

    def build(self):
        # Call UnitModel.build to setup dynamics
        super().build()

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        units_meta = self.config.property_package.get_metadata().get_derived_units

        # Check configs for errors
        self._validate_config()
        self._process_config()

        if hasattr(self.config.property_package, "ion_set"):
            solute_set = self.config.property_package.ion_set
        elif hasattr(self.config.property_package, "solute_set"):
            solute_set = self.config.property_package.solute_set

        solvent_solute_set = self.config.property_package.solvent_set | solute_set

        # Add unit parameters
        self.flux_vol_solvent = Var(
            self.flowsheet().config.time,
            self.config.property_package.solvent_set,
            initialize=1.67e-6,
            bounds=(0.0, 1e-4),
            units=units_meta("length") * units_meta("time") ** -1,
            doc="Solvent volumetric flux",
        )
        self.rejection_phase_comp = Var(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            solute_set,
            initialize=0.9,
            bounds=(-1 + 1e-6, 1 - 1e-6),
            units=pyunits.dimensionless,
            doc="Observed solute rejection",
        )
        self.dens_solvent = Param(
            initialize=1000,
            units=units_meta("mass") * units_meta("length") ** -3,
            doc="Pure water density",
        )

        # Add unit variables
        self.area = Var(
            initialize=1,
            bounds=(0.0, 1e6),
            domain=NonNegativeReals,
            units=units_meta("length") ** 2,
            doc="Membrane area",
        )

        def recovery_mass_phase_comp_initialize(b, t, p, j):
            if j in b.config.property_package.solvent_set:
                return 0.8
            elif j in solute_set:
                return 0.1

        def recovery_mass_phase_comp_bounds(b, t, p, j):
            ub = 1 - 1e-6
            if j in b.config.property_package.solvent_set:
                lb = 1e-2
            elif j in solute_set:
                lb = 1e-5
            else:
                lb = 1e-5

            return lb, ub

        self.recovery_mass_phase_comp = Var(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            solvent_solute_set,
            initialize=recovery_mass_phase_comp_initialize,
            bounds=recovery_mass_phase_comp_bounds,
            units=pyunits.dimensionless,
            doc="Mass-based component recovery",
        )
        self.recovery_vol_phase = Var(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            initialize=0.1,
            bounds=(1e-2, 1 - 1e-6),
            units=pyunits.dimensionless,
            doc="Volumetric-based recovery",
        )

        # Build control volume for feed side
        self.feed_side = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.feed_side.add_state_blocks(has_phase_equilibrium=False)

        self.feed_side.add_material_balances(
            balance_type=self.config.material_balance_type, has_mass_transfer=True
        )

        self.feed_side.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_enthalpy_transfer=False,
        )

        if self.config.is_isothermal:
            self.feed_side.add_isothermal_assumption()

        self.feed_side.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change,
        )

        # Add permeate block
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = False  # permeate block is not an inlet
        self.properties_permeate = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of permeate",
            **tmp_dict,
        )

        # Add Ports
        self.add_inlet_port(name="inlet", block=self.feed_side)
        self.add_outlet_port(name="retentate", block=self.feed_side)
        self.add_port(name="permeate", block=self.properties_permeate)

        # References for control volume
        # pressure change
        if (
            self.config.has_pressure_change is True
            and self.config.momentum_balance_type != "none"
        ):
            self.deltaP = Reference(self.feed_side.deltaP)

        # mass transfer
        self.mass_transfer_phase_comp = Var(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            solvent_solute_set,
            initialize=1,
            bounds=(0.0, 1e6),
            domain=NonNegativeReals,
            units=units_meta("mass") * units_meta("time") ** -1,
            doc="Mass transfer to permeate",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            solvent_solute_set,
            doc="Mass transfer term",
        )
        def eq_mass_transfer_term(b, t, p, j):
            # TODO- come up with better way to handle different locations of mw_comp in property models (generic vs simple ion prop model)
            if (
                b.feed_side.properties_in[0].get_material_flow_basis()
                == MaterialFlowBasis.mass
            ):
                return (
                    b.mass_transfer_phase_comp[t, p, j]
                    == -b.feed_side.mass_transfer_term[t, p, j]
                )
            elif (
                b.feed_side.properties_in[0].get_material_flow_basis()
                == MaterialFlowBasis.molar
            ):
                if hasattr(b.feed_side.properties_in[0].params, "mw_comp"):
                    mw_comp = b.feed_side.properties_in[0].params.mw_comp[j]
                elif hasattr(b.feed_side.properties_in[0], "mw_comp"):
                    mw_comp = b.feed_side.properties_in[0].mw_comp[j]
                else:
                    raise ConfigurationError("mw_comp was not found.")
                return (
                    b.mass_transfer_phase_comp[t, p, j]
                    == -b.feed_side.mass_transfer_term[t, p, j] * mw_comp
                )

        # NF performance equations
        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.solvent_set,
            doc="Solvent mass transfer",
        )
        def eq_solvent_transfer(b, t, p, j):
            if (
                b.feed_side.properties_in[0].get_material_flow_basis()
                == MaterialFlowBasis.mass
            ):
                return (
                    b.flux_vol_solvent[t, j] * b.dens_solvent * b.area
                    == -b.feed_side.mass_transfer_term[t, p, j]
                )
            elif (
                b.feed_side.properties_in[0].get_material_flow_basis()
                == MaterialFlowBasis.molar
            ):
                # TODO- come up with better way to handle different locations of mw_comp in property models (generic vs simple ion prop model)
                if hasattr(b.feed_side.properties_in[0].params, "mw_comp"):
                    mw_comp = b.feed_side.properties_in[0].params.mw_comp[j]
                elif hasattr(b.feed_side.properties_in[0], "mw_comp"):
                    mw_comp = b.feed_side.properties_in[0].mw_comp[j]
                else:
                    raise ConfigurationError("mw_comp was not found.")
                return (
                    b.flux_vol_solvent[t, j] * b.dens_solvent * b.area
                    == -b.feed_side.mass_transfer_term[t, p, j] * mw_comp
                )

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            solvent_solute_set,
            doc="Permeate production",
        )
        def eq_permeate_production(b, t, p, j):
            return (
                b.properties_permeate[t].get_material_flow_terms(p, j)
                == -b.feed_side.mass_transfer_term[t, p, j]
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            solute_set,
            doc="Solute rejection",
        )
        def eq_rejection_phase_comp(b, t, p, j):
            return b.properties_permeate[t].conc_mol_phase_comp[
                "Liq", j
            ] == b.feed_side.properties_in[t].conc_mol_phase_comp["Liq", j] * (
                1 - b.rejection_phase_comp[t, p, j]
            )

        @self.Constraint(self.flowsheet().config.time)
        def eq_recovery_vol_phase(b, t):
            return (
                b.recovery_vol_phase[t, "Liq"]
                == b.properties_permeate[t].flow_vol
                / b.feed_side.properties_in[t].flow_vol
            )

        @self.Constraint(self.flowsheet().config.time, solvent_solute_set)
        def eq_recovery_mass_phase_comp(b, t, j):
            return (
                b.recovery_mass_phase_comp[t, "Liq", j]
                == b.properties_permeate[t].flow_mass_phase_comp["Liq", j]
                / b.feed_side.properties_in[t].flow_mass_phase_comp["Liq", j]
            )

        @self.Constraint(
            self.flowsheet().config.time, doc="Isothermal assumption for permeate"
        )
        def eq_permeate_isothermal(b, t):
            return (
                b.feed_side.properties_in[t].temperature
                == b.properties_permeate[t].temperature
            )

    def initialize_build(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize holdup block
        flags = self.feed_side.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )
        init_log.info_high("Initialization Step 1 Complete.")
        # ---------------------------------------------------------------------
        # Initialize permeate
        # Set state_args from inlet state
        if state_args is None:
            state_args = {}
            state_dict = self.feed_side.properties_in[
                self.flowsheet().config.time.first()
            ].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        self.properties_permeate.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )
        init_log.info_high("Initialization Step 2 Complete.")

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release Inlet state
        self.feed_side.release_state(flags, outlvl + 1)
        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

    def _get_performance_contents(self, time_point=0):
        for k in ("ion_set", "solute_set"):
            if hasattr(self.config.property_package, k):
                solute_set = getattr(self.config.property_package, k)
                break
        var_dict = {}
        expr_dict = {}
        var_dict["Volumetric Recovery Rate"] = self.recovery_vol_phase[
            time_point, "Liq"
        ]
        var_dict["Solvent Mass Recovery Rate"] = self.recovery_mass_phase_comp[
            time_point, "Liq", "H2O"
        ]
        var_dict["Membrane Area"] = self.area
        if hasattr(self, "deltaP"):
            var_dict["Pressure Change"] = self.deltaP[time_point]
        if self.feed_side.properties_in[time_point].is_property_constructed("flow_vol"):
            if self.feed_side.properties_in[time_point].flow_vol.is_variable_type():
                obj_dict = var_dict
            elif self.feed_side.properties_in[
                time_point
            ].flow_vol.is_named_expression_type():
                obj_dict = expr_dict
            else:
                raise Exception(
                    f"{self.feed_side.properties_in[time_point].flow_vol} isn't a variable nor expression"
                )
            obj_dict["Volumetric Flowrate @Inlet"] = self.feed_side.properties_in[
                time_point
            ].flow_vol
        if self.feed_side.properties_out[time_point].is_property_constructed(
            "flow_vol"
        ):
            if self.feed_side.properties_out[time_point].flow_vol.is_variable_type():
                obj_dict = var_dict
            elif self.feed_side.properties_out[
                time_point
            ].flow_vol.is_named_expression_type():
                obj_dict = expr_dict
            else:
                raise Exception(
                    f"{self.feed_side.properties_in[time_point].flow_vol} isn't a variable nor expression"
                )
            obj_dict["Volumetric Flowrate @Outlet"] = self.feed_side.properties_out[
                time_point
            ].flow_vol
        var_dict["Solvent Volumetric Flux"] = self.flux_vol_solvent[time_point, "H2O"]
        for j in solute_set:
            var_dict[f"{j} Rejection"] = self.rejection_phase_comp[time_point, "Liq", j]
            if (
                self.feed_side.properties_in[time_point]
                .conc_mol_phase_comp["Liq", j]
                .is_expression_type()
            ):
                obj_dict = expr_dict
            elif (
                self.feed_side.properties_in[time_point]
                .conc_mol_phase_comp["Liq", j]
                .is_variable_type()
            ):
                obj_dict = var_dict
            obj_dict[f"{j} Molar Concentration @Inlet"] = self.feed_side.properties_in[
                time_point
            ].conc_mol_phase_comp["Liq", j]
            obj_dict[
                f"{j} Molar Concentration @Outlet"
            ] = self.feed_side.properties_out[time_point].conc_mol_phase_comp["Liq", j]
            obj_dict[f"{j} Molar Concentration @Permeate"] = self.properties_permeate[
                time_point
            ].conc_mol_phase_comp["Liq", j]

        return {"vars": var_dict, "exprs": expr_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Feed Inlet": self.inlet,
                "Feed Outlet": self.retentate,
                "Permeate Outlet": self.permeate,
            },
            time_point=time_point,
        )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        for k in ("ion_set", "solute_set"):
            if hasattr(self.config.property_package, k):
                solute_set = getattr(self.config.property_package, k)
                break

        # TODO: require users to set scaling factor for area or calculate it based on mass transfer and flux
        iscale.set_scaling_factor(self.area, 1e-1)

        # setting scaling factors for variables
        # these variables should have user input, if not there will be a warning
        if iscale.get_scaling_factor(self.area) is None:
            sf = iscale.get_scaling_factor(self.area, default=1, warning=True)
            iscale.set_scaling_factor(self.area, sf)

        # these variables do not typically require user input,
        # will not override if the user does provide the scaling factor
        # TODO: this default scaling assumes SI units rather than being based on the property package
        if iscale.get_scaling_factor(self.dens_solvent) is None:
            iscale.set_scaling_factor(self.dens_solvent, 1e-3)

        for t, v in self.flux_vol_solvent.items():
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, 1e6)

        for (t, p, j), v in self.rejection_phase_comp.items():
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, 1e1)

        for (t, p, j), v in self.mass_transfer_phase_comp.items():
            if iscale.get_scaling_factor(v) is None:
                sf = 10 * iscale.get_scaling_factor(
                    self.feed_side.properties_in[t].get_material_flow_terms(p, j)
                )
                iscale.set_scaling_factor(v, sf)
        if iscale.get_scaling_factor(self.recovery_vol_phase) is None:
            iscale.set_scaling_factor(self.recovery_vol_phase, 1)

        for (t, p, j), v in self.recovery_mass_phase_comp.items():
            if j in self.config.property_package.solvent_set:
                sf = 1
            elif j in solute_set:
                sf = 10
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)

        # transforming constraints

        for ind, c in self.eq_mass_transfer_term.items():
            sf = iscale.get_scaling_factor(self.mass_transfer_phase_comp[ind])
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_solvent_transfer.items():
            sf = iscale.get_scaling_factor(self.mass_transfer_phase_comp[ind])
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_permeate_production.items():
            sf = iscale.get_scaling_factor(self.mass_transfer_phase_comp[ind])
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_rejection_phase_comp.items():
            sf = iscale.get_scaling_factor(self.rejection_phase_comp[ind])
            iscale.constraint_scaling_transform(c, sf)

        for t, c in self.eq_permeate_isothermal.items():
            sf = iscale.get_scaling_factor(self.feed_side.properties_in[t].temperature)
            iscale.constraint_scaling_transform(c, sf)

        for t, c in self.eq_recovery_vol_phase.items():
            sf = iscale.get_scaling_factor(self.recovery_vol_phase[t, "Liq"])
            iscale.constraint_scaling_transform(c, sf)

        for (t, j), c in self.eq_recovery_mass_phase_comp.items():
            sf = iscale.get_scaling_factor(self.recovery_mass_phase_comp[t, "Liq", j])
            iscale.constraint_scaling_transform(c, sf)

    @property
    def default_costing_method(self):
        return cost_nanofiltration



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest

# Imports from idaes core
from idaes.core import AqueousPhase
from idaes.core.base.components import Solvent, Solute, Cation, Anion
from idaes.core.base.phases import PhaseType as PT

# Imports from idaes generic models
from idaes.models.properties.modular_properties.pure.ConstantProperties import Constant
from idaes.models.properties.modular_properties.state_definitions import FpcTP
from idaes.models.properties.modular_properties.eos.ideal import Ideal

# Importing the enum for concentration unit basis used in the 'get_concentration_term' function
from idaes.models.properties.modular_properties.base.generic_reaction import (
    ConcentrationForm,
)

# Import the object/function for heat of reaction
from idaes.models.properties.modular_properties.reactions.dh_rxn import constant_dh_rxn

# Import safe log power law equation
from idaes.models.properties.modular_properties.reactions.equilibrium_forms import (
    log_power_law_equil,
)

# Import k-value functions
from idaes.models.properties.modular_properties.reactions.equilibrium_constant import (
    van_t_hoff,
)

# Import the idaes objects for Generic Properties and Reactions
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)

from watertap.property_models.multicomp_aq_sol_prop_pack import MCASParameterBlock

from watertap.unit_models.boron_removal import BoronRemoval
from pyomo.environ import (
    ConcreteModel,
    assert_optimal_termination,
    value,
    Param,
    Var,
    units as pyunits,
    Constraint,
)
from idaes.core import (
    FlowsheetBlock,
)
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.model_statistics import degrees_of_freedom
from pyomo.util.check_units import assert_units_consistent
import idaes.core.util.scaling as iscale
from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver
import re

__author__ = "Austin Ladshaw"

solver = get_solver()

# Helper function for multiple test setup
def model_setup(
    m,
    chem_add=1.8e-5,
    state={
        "H2O": 100,
        "H_+": 1e-7,
        "OH_-": 1e-7,
        "B[OH]3": 2e-4,
        "B[OH]4_-": 1e-6,
        "Na_+": 1e-4,
        "HCO3_-": 1e-4,
    },
):

    m.fs.unit.inlet.pressure.fix(101325)
    m.fs.unit.inlet.temperature.fix(298.15)
    for j in state:
        idx = (0, "Liq", j)
        if idx in m.fs.unit.inlet.flow_mol_phase_comp:
            m.fs.unit.inlet.flow_mol_phase_comp[idx].fix(state[j])
    m.fs.unit.caustic_dose_rate.fix(chem_add)
    m.fs.unit.reactor_volume.fix(1)


# Helper function to automate scaling
def scaling_setup(
    m,
    state={
        "H2O": 100,
        "H_+": 5e-5,
        "OH_-": 5e-5,
        "B[OH]3": 2e-4,
        "B[OH]4_-": 1e-6,
        "Na_+": 1e-4,
        "HCO3_-": 1e-4,
    },
):
    # Set some scaling factors and look for 'bad' scaling
    for j in state:
        idx = (0, "Liq", j)
        if idx in m.fs.unit.inlet.flow_mol_phase_comp:
            m.fs.properties.set_default_scaling(
                "flow_mol_phase_comp", 1 / state[j], index=("Liq", j)
            )

    iscale.calculate_scaling_factors(m.fs)


# -----------------------------------------------------------------------------
# Start test class
class TestBoronRemoval_IonPropPack_Min:
    @pytest.fixture(scope="class")
    def min_boron_removal_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        # create dict to define ions (the prop pack requires this)
        ion_dict = {
            "solute_list": ["B[OH]3", "B[OH]4_-", "Na_+"],
            "mw_data": {
                "H2O": 18e-3,
                "B[OH]3": 61.83e-3,
                "B[OH]4_-": 78.83e-3,
                "Na_+": 23e-3,
            },
            "charge": {
                "B[OH]4_-": -1,
                "Na_+": 1,
                "B[OH]3": 0,
            },
        }

        # attach prop pack to flowsheet
        m.fs.properties = MCASParameterBlock(**ion_dict)

        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "caustic_additive": {
                "cation_name": "Na_+",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        m.fs.unit = BoronRemoval(
            property_package=m.fs.properties, chemical_mapping_data=map
        )

        return m

    @pytest.mark.unit
    def test_build_model(self, min_boron_removal_model):
        m = min_boron_removal_model

        assert isinstance(m.fs.unit.ion_charge, Param)
        assert len(m.fs.unit.ion_charge) == 2

        assert hasattr(m.fs.unit, "boron_name_id")
        assert hasattr(m.fs.unit, "borate_name_id")
        assert hasattr(m.fs.unit, "proton_name_id")
        assert hasattr(m.fs.unit, "hydroxide_name_id")
        assert hasattr(m.fs.unit, "cation_name_id")

        assert hasattr(m.fs.unit, "caustic_chem_name")

        assert hasattr(m.fs.unit, "control_volume")

        assert isinstance(m.fs.unit.caustic_mw, Param)
        assert isinstance(m.fs.unit.additive_molar_ratio, Param)
        assert isinstance(m.fs.unit.caustic_dose_rate, Var)

        assert isinstance(m.fs.unit.Kw_0, Param)
        assert isinstance(m.fs.unit.dH_w, Param)
        assert isinstance(m.fs.unit.Ka_0, Param)
        assert isinstance(m.fs.unit.dH_a, Param)

        assert isinstance(m.fs.unit.conc_mol_H, Var)
        assert isinstance(m.fs.unit.conc_mol_OH, Var)
        assert isinstance(m.fs.unit.conc_mol_Boron, Var)
        assert isinstance(m.fs.unit.conc_mol_Borate, Var)

        assert isinstance(m.fs.unit.eq_mass_transfer_term, Constraint)
        assert isinstance(m.fs.unit.eq_electroneutrality, Constraint)
        assert isinstance(m.fs.unit.eq_total_boron, Constraint)
        assert isinstance(m.fs.unit.eq_water_dissociation, Constraint)
        assert isinstance(m.fs.unit.eq_boron_dissociation, Constraint)

    @pytest.mark.unit
    def test_stats(self, min_boron_removal_model):
        m = min_boron_removal_model

        # Check to make sure we have the correct DOF and
        #   check to make sure the units are correct
        assert_units_consistent(m)
        assert degrees_of_freedom(m) == 8

        # set the variables
        state = {
            "H2O": 100,
            "H_+": 1e-7,
            "OH_-": 1e-7,
            "B[OH]3": 2e-4,
            "B[OH]4_-": 1e-6,
            "Na_+": 1e-15,
            "HCO3_-": 1e-4,
        }
        # For this test, we don't actually know how much
        #   of the base to add. Instead, we have a target
        #   exit flow of boron (see below). We will initially
        #   guess that we want 5 mg/L of additive, but the
        #   actual solution is 10 mg/L.
        model_setup(m, chem_add=0.9e-5, state=state)

        # Modified this test to fix the desired outlet flow
        #   of Boron and unfix the dosage needed to get that outlet
        m.fs.unit.outlet.flow_mol_phase_comp[(0, "Liq", "B[OH]3")].fix(1.98677e-5)
        m.fs.unit.caustic_dose_rate.unfix()

        assert degrees_of_freedom(m) == 0

    @pytest.mark.unit
    def test_scaling(self, min_boron_removal_model):
        m = min_boron_removal_model

        # Set some scaling factors and look for 'bad' scaling
        scaling_setup(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(iscale.unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0

        # check if any variables are badly scaled
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                m, large=1e3, small=1e-3
            )
        }
        assert not badly_scaled_var_values

    @pytest.mark.component
    def test_initialization(self, min_boron_removal_model):
        m = min_boron_removal_model
        initialization_tester(m)

        # check to make sure DOF does not change
        assert degrees_of_freedom(m) == 0

    @pytest.mark.component
    def test_solve(self, min_boron_removal_model):
        m = min_boron_removal_model

        # first, check to make sure that after initialized, the scaling is still good
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                m, large=1e3, small=1e-3
            )
        }
        assert not badly_scaled_var_values

        # run solver and check for optimal solution
        results = solver.solve(m, tee=True)
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, min_boron_removal_model):
        m = min_boron_removal_model

        assert value(
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", m.fs.unit.boron_name_id]
        ) == pytest.approx(1.98677e-5, rel=1e-4)
        assert value(
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", m.fs.unit.borate_name_id]
        ) == pytest.approx(1.81133e-4, rel=1e-4)
        assert value(m.fs.unit.outlet_pH()) == pytest.approx(10.171, rel=1e-4)
        assert value(m.fs.unit.outlet_pOH()) == pytest.approx(3.8257, rel=1e-4)
        assert value(m.fs.unit.caustic_dose_rate[0].value) == pytest.approx(
            1.8e-5, rel=1e-4
        )


# -----------------------------------------------------------------------------
# Start test class
class TestBoronRemoval_IonPropPack_with_ResAlk:
    @pytest.fixture(scope="class")
    def alk_boron_removal_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # create dict to define ions (the prop pack requires this)
        ion_dict = {
            "solute_list": ["B[OH]3", "B[OH]4_-", "Na_+", "HCO3_-"],
            "mw_data": {
                "H2O": 18e-3,
                "B[OH]3": 61.83e-3,
                "B[OH]4_-": 78.83e-3,
                "Na_+": 23e-3,
                "HCO3_-": 61e-3,
            },
            "charge": {"B[OH]3": 0, "B[OH]4_-": -1, "Na_+": 1, "HCO3_-": -1},
        }

        # attach prop pack to flowsheet
        m.fs.properties = MCASParameterBlock(**ion_dict)

        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "caustic_additive": {
                "cation_name": "Na_+",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        m.fs.unit = BoronRemoval(
            property_package=m.fs.properties, chemical_mapping_data=map
        )

        return m

    @pytest.mark.unit
    def test_build_model(self, alk_boron_removal_model):
        m = alk_boron_removal_model

        assert isinstance(m.fs.unit.ion_charge, Param)
        assert len(m.fs.unit.ion_charge) == 3

        assert hasattr(m.fs.unit, "boron_name_id")
        assert hasattr(m.fs.unit, "borate_name_id")
        assert hasattr(m.fs.unit, "proton_name_id")
        assert hasattr(m.fs.unit, "hydroxide_name_id")
        assert hasattr(m.fs.unit, "cation_name_id")

        assert hasattr(m.fs.unit, "caustic_chem_name")

        assert hasattr(m.fs.unit, "control_volume")

        assert isinstance(m.fs.unit.caustic_mw, Param)
        assert isinstance(m.fs.unit.additive_molar_ratio, Param)
        assert isinstance(m.fs.unit.caustic_dose_rate, Var)

        assert isinstance(m.fs.unit.Kw_0, Param)
        assert isinstance(m.fs.unit.dH_w, Param)
        assert isinstance(m.fs.unit.Ka_0, Param)
        assert isinstance(m.fs.unit.dH_a, Param)

        assert isinstance(m.fs.unit.conc_mol_H, Var)
        assert isinstance(m.fs.unit.conc_mol_OH, Var)
        assert isinstance(m.fs.unit.conc_mol_Boron, Var)
        assert isinstance(m.fs.unit.conc_mol_Borate, Var)

        assert isinstance(m.fs.unit.eq_mass_transfer_term, Constraint)
        assert isinstance(m.fs.unit.eq_electroneutrality, Constraint)
        assert isinstance(m.fs.unit.eq_total_boron, Constraint)
        assert isinstance(m.fs.unit.eq_water_dissociation, Constraint)
        assert isinstance(m.fs.unit.eq_boron_dissociation, Constraint)

    @pytest.mark.unit
    def test_stats(self, alk_boron_removal_model):
        m = alk_boron_removal_model

        # Check to make sure we have the correct DOF and
        #   check to make sure the units are correct
        assert_units_consistent(m)
        assert degrees_of_freedom(m) == 9

        # set the variables
        model_setup(m, chem_add=1.8e-4)

        assert degrees_of_freedom(m) == 0

    @pytest.mark.unit
    def test_scaling(self, alk_boron_removal_model):
        m = alk_boron_removal_model

        # Set some scaling factors and look for 'bad' scaling
        scaling_setup(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(iscale.unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0

        # check if any variables are badly scaled
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                m, large=1e3, small=1e-3
            )
        }
        assert not badly_scaled_var_values

    @pytest.mark.component
    def test_initialization(self, alk_boron_removal_model):
        m = alk_boron_removal_model
        initialization_tester(m)

        # check to make sure DOF does not change
        assert degrees_of_freedom(m) == 0

    @pytest.mark.component
    def test_solve(self, alk_boron_removal_model):
        m = alk_boron_removal_model

        # first, check to make sure that after initialized, the scaling is still good
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                m, large=1e3, small=1e-3
            )
        }
        assert not badly_scaled_var_values

        # run solver and check for optimal solution
        results = solver.solve(m)
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, alk_boron_removal_model):
        m = alk_boron_removal_model

        assert value(
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", m.fs.unit.boron_name_id]
        ) == pytest.approx(1.36911e-6, rel=1e-4)
        assert value(
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", m.fs.unit.borate_name_id]
        ) == pytest.approx(1.99642e-4, rel=1e-4)
        assert value(m.fs.unit.outlet_pH()) == pytest.approx(11.375, rel=1e-4)
        assert value(m.fs.unit.outlet_pOH()) == pytest.approx(2.6218, rel=1e-4)


# -----------------------------------------------------------------------------
# Start test class
class TestBoronRemoval_IonPropPack_with_ResBase:
    @pytest.fixture(scope="class")
    def base_boron_removal_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # create dict to define ions (the prop pack requires this)
        ion_dict = {
            "solute_list": ["B[OH]3", "B[OH]4_-", "Na_+"],
            "mw_data": {
                "H2O": 18e-3,
                "B[OH]3": 61.83e-3,
                "B[OH]4_-": 78.83e-3,
                "Na_+": 23e-3,
            },
            "charge": {
                "B[OH]3": 0,
                "B[OH]4_-": -1,
                "Na_+": 1,
            },
        }

        # attach prop pack to flowsheet
        m.fs.properties = MCASParameterBlock(**ion_dict)

        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "caustic_additive": {
                "cation_name": "Na_+",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        m.fs.unit = BoronRemoval(
            property_package=m.fs.properties, chemical_mapping_data=map
        )

        return m

    @pytest.mark.unit
    def test_build_model(self, base_boron_removal_model):
        m = base_boron_removal_model

        assert isinstance(m.fs.unit.ion_charge, Param)
        assert len(m.fs.unit.ion_charge) == 2

        assert hasattr(m.fs.unit, "boron_name_id")
        assert hasattr(m.fs.unit, "borate_name_id")
        assert hasattr(m.fs.unit, "proton_name_id")
        assert hasattr(m.fs.unit, "hydroxide_name_id")
        assert hasattr(m.fs.unit, "cation_name_id")

        assert hasattr(m.fs.unit, "caustic_chem_name")

        assert hasattr(m.fs.unit, "control_volume")

        assert isinstance(m.fs.unit.caustic_mw, Param)
        assert isinstance(m.fs.unit.additive_molar_ratio, Param)
        assert isinstance(m.fs.unit.caustic_dose_rate, Var)

        assert isinstance(m.fs.unit.Kw_0, Param)
        assert isinstance(m.fs.unit.dH_w, Param)
        assert isinstance(m.fs.unit.Ka_0, Param)
        assert isinstance(m.fs.unit.dH_a, Param)

        assert isinstance(m.fs.unit.conc_mol_H, Var)
        assert isinstance(m.fs.unit.conc_mol_OH, Var)
        assert isinstance(m.fs.unit.conc_mol_Boron, Var)
        assert isinstance(m.fs.unit.conc_mol_Borate, Var)

        assert isinstance(m.fs.unit.eq_mass_transfer_term, Constraint)
        assert isinstance(m.fs.unit.eq_electroneutrality, Constraint)
        assert isinstance(m.fs.unit.eq_total_boron, Constraint)
        assert isinstance(m.fs.unit.eq_water_dissociation, Constraint)
        assert isinstance(m.fs.unit.eq_boron_dissociation, Constraint)

    @pytest.mark.unit
    def test_stats(self, base_boron_removal_model):
        m = base_boron_removal_model

        # Check to make sure we have the correct DOF and
        #   check to make sure the units are correct
        assert_units_consistent(m)
        assert degrees_of_freedom(m) == 8

        # set the variables
        model_setup(m, chem_add=0)

        assert degrees_of_freedom(m) == 0

    @pytest.mark.unit
    def test_scaling(self, base_boron_removal_model):
        m = base_boron_removal_model

        # Set some scaling factors and look for 'bad' scaling
        scaling_setup(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(iscale.unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0

        # check if any variables are badly scaled
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                m, large=1e3, small=1e-3
            )
        }
        assert not badly_scaled_var_values

    @pytest.mark.component
    def test_initialization(self, base_boron_removal_model):
        m = base_boron_removal_model
        initialization_tester(m)

        # check to make sure DOF does not change
        assert degrees_of_freedom(m) == 0

    @pytest.mark.component
    def test_solve(self, base_boron_removal_model):
        m = base_boron_removal_model

        # first, check to make sure that after initialized, the scaling is still good
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                m, large=1e3, small=1e-3
            )
        }
        assert not badly_scaled_var_values

        # run solver and check for optimal solution
        results = solver.solve(m)
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, base_boron_removal_model):
        m = base_boron_removal_model

        assert value(
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", m.fs.unit.boron_name_id]
        ) == pytest.approx(1.20642e-4, rel=1e-4)
        assert value(
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", m.fs.unit.borate_name_id]
        ) == pytest.approx(8.03579e-5, rel=1e-4)
        assert value(m.fs.unit.outlet_pH()) == pytest.approx(9.0343, rel=1e-4)
        assert value(m.fs.unit.outlet_pOH()) == pytest.approx(4.9621, rel=1e-4)


# -----------------------------------------------------------------------------
# Start test class
class TestBoronRemoval_GenPropPack:
    @pytest.fixture(scope="class")
    def gen_boron_removal_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Configuration dictionary for generic
        thermo_config = {
            "components": {
                "H2O": {
                    "type": Solvent,
                    # Define the methods used to calculate the following properties
                    "dens_mol_liq_comp": Constant,
                    "enth_mol_liq_comp": Constant,
                    "cp_mol_liq_comp": Constant,
                    "entr_mol_liq_comp": Constant,
                    # Parameter data is always associated with the methods defined above
                    "parameter_data": {
                        "mw": (18.0153, pyunits.g / pyunits.mol),
                        "dens_mol_liq_comp_coeff": (
                            55.2,
                            pyunits.kmol * pyunits.m**-3,
                        ),
                        "cp_mol_liq_comp_coeff": (
                            75.312,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_liq_comp_ref": (0, pyunits.kJ / pyunits.mol),
                        "entr_mol_form_liq_comp_ref": (
                            0,
                            pyunits.J / pyunits.K / pyunits.mol,
                        ),
                    },
                    # End parameter_data
                },
                "H_+": {
                    "type": Cation,
                    "charge": 1,
                    # Define the methods used to calculate the following properties
                    "dens_mol_liq_comp": Constant,
                    "enth_mol_liq_comp": Constant,
                    "cp_mol_liq_comp": Constant,
                    "entr_mol_liq_comp": Constant,
                    # Parameter data is always associated with the methods defined above
                    "parameter_data": {
                        "mw": (1, pyunits.g / pyunits.mol),
                        "dens_mol_liq_comp_coeff": (
                            55.2,
                            pyunits.kmol * pyunits.m**-3,
                        ),
                        "cp_mol_liq_comp_coeff": (
                            75.312,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_liq_comp_ref": (0, pyunits.kJ / pyunits.mol),
                        "entr_mol_form_liq_comp_ref": (
                            0,
                            pyunits.J / pyunits.K / pyunits.mol,
                        ),
                    },
                    # End parameter_data
                },
                "Na_+": {
                    "type": Cation,
                    "charge": 1,
                    # Define the methods used to calculate the following properties
                    "dens_mol_liq_comp": Constant,
                    "enth_mol_liq_comp": Constant,
                    "cp_mol_liq_comp": Constant,
                    "entr_mol_liq_comp": Constant,
                    # Parameter data is always associated with the methods defined above
                    "parameter_data": {
                        "mw": (23, pyunits.g / pyunits.mol),
                        "dens_mol_liq_comp_coeff": (
                            55.2,
                            pyunits.kmol * pyunits.m**-3,
                        ),
                        "cp_mol_liq_comp_coeff": (
                            75.312,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_liq_comp_ref": (0, pyunits.kJ / pyunits.mol),
                        "entr_mol_form_liq_comp_ref": (
                            0,
                            pyunits.J / pyunits.K / pyunits.mol,
                        ),
                    },
                    # End parameter_data
                },
                "OH_-": {
                    "type": Anion,
                    "charge": -1,
                    # Define the methods used to calculate the following properties
                    "dens_mol_liq_comp": Constant,
                    "enth_mol_liq_comp": Constant,
                    "cp_mol_liq_comp": Constant,
                    "entr_mol_liq_comp": Constant,
                    # Parameter data is always associated with the methods defined above
                    "parameter_data": {
                        "mw": (17, pyunits.g / pyunits.mol),
                        "dens_mol_liq_comp_coeff": (
                            55.2,
                            pyunits.kmol * pyunits.m**-3,
                        ),
                        "cp_mol_liq_comp_coeff": (
                            75.312,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_liq_comp_ref": (0, pyunits.kJ / pyunits.mol),
                        "entr_mol_form_liq_comp_ref": (
                            0,
                            pyunits.J / pyunits.K / pyunits.mol,
                        ),
                    },
                    # End parameter_data
                },
                "B[OH]4_-": {
                    "type": Anion,
                    "charge": -1,
                    # Define the methods used to calculate the following properties
                    "dens_mol_liq_comp": Constant,
                    "enth_mol_liq_comp": Constant,
                    "cp_mol_liq_comp": Constant,
                    "entr_mol_liq_comp": Constant,
                    # Parameter data is always associated with the methods defined above
                    "parameter_data": {
                        "mw": (78.83, pyunits.g / pyunits.mol),
                        "dens_mol_liq_comp_coeff": (
                            55.2,
                            pyunits.kmol * pyunits.m**-3,
                        ),
                        "cp_mol_liq_comp_coeff": (
                            75.312,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_liq_comp_ref": (0, pyunits.kJ / pyunits.mol),
                        "entr_mol_form_liq_comp_ref": (
                            0,
                            pyunits.J / pyunits.K / pyunits.mol,
                        ),
                    },
                    # End parameter_data
                },
                "B[OH]3": {
                    "type": Solute,
                    "valid_phase_types": PT.aqueousPhase,
                    # Define the methods used to calculate the following properties
                    "dens_mol_liq_comp": Constant,
                    "enth_mol_liq_comp": Constant,
                    "cp_mol_liq_comp": Constant,
                    "entr_mol_liq_comp": Constant,
                    # Parameter data is always associated with the methods defined above
                    "parameter_data": {
                        "mw": (61.83, pyunits.g / pyunits.mol),
                        "dens_mol_liq_comp_coeff": (
                            55.2,
                            pyunits.kmol * pyunits.m**-3,
                        ),
                        "cp_mol_liq_comp_coeff": (
                            75.312,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_liq_comp_ref": (0, pyunits.kJ / pyunits.mol),
                        "entr_mol_form_liq_comp_ref": (
                            0,
                            pyunits.J / pyunits.K / pyunits.mol,
                        ),
                    },
                    # End parameter_data
                },
            },
            # End Component list
            "phases": {
                "Liq": {"type": AqueousPhase, "equation_of_state": Ideal},
            },
            "state_definition": FpcTP,
            "state_bounds": {
                "temperature": (273.15, 300, 650),
                "pressure": (5e4, 1e5, 1e6),
            },
            "pressure_ref": 1e5,
            "temperature_ref": 300,
            "base_units": {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
        }
        # End thermo_config definition

        # attach prop pack to flowsheet
        m.fs.properties = GenericParameterBlock(**thermo_config)

        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "proton_name": "H_+",  # [is optional]
            "hydroxide_name": "OH_-",  # [is optional]
            "caustic_additive": {
                "cation_name": "Na_+",  # [is required]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        m.fs.unit = BoronRemoval(
            property_package=m.fs.properties, chemical_mapping_data=map
        )

        return m

    @pytest.mark.unit
    def test_build_model(self, gen_boron_removal_model):
        m = gen_boron_removal_model

        assert isinstance(m.fs.unit.ion_charge, Param)
        assert len(m.fs.unit.ion_charge) == 4

        assert hasattr(m.fs.unit, "boron_name_id")
        assert hasattr(m.fs.unit, "borate_name_id")
        assert hasattr(m.fs.unit, "proton_name_id")
        assert hasattr(m.fs.unit, "hydroxide_name_id")
        assert hasattr(m.fs.unit, "cation_name_id")

        assert hasattr(m.fs.unit, "caustic_chem_name")

        assert hasattr(m.fs.unit, "control_volume")

        assert isinstance(m.fs.unit.caustic_mw, Param)
        assert isinstance(m.fs.unit.additive_molar_ratio, Param)
        assert isinstance(m.fs.unit.caustic_dose_rate, Var)

        assert isinstance(m.fs.unit.Kw_0, Param)
        assert isinstance(m.fs.unit.dH_w, Param)
        assert isinstance(m.fs.unit.Ka_0, Param)
        assert isinstance(m.fs.unit.dH_a, Param)

        assert isinstance(m.fs.unit.conc_mol_H, Var)
        assert isinstance(m.fs.unit.conc_mol_OH, Var)
        assert isinstance(m.fs.unit.conc_mol_Boron, Var)
        assert isinstance(m.fs.unit.conc_mol_Borate, Var)

        assert isinstance(m.fs.unit.eq_mass_transfer_term, Constraint)
        assert isinstance(m.fs.unit.eq_electroneutrality, Constraint)
        assert isinstance(m.fs.unit.eq_total_boron, Constraint)
        assert isinstance(m.fs.unit.eq_water_dissociation, Constraint)
        assert isinstance(m.fs.unit.eq_boron_dissociation, Constraint)

    @pytest.mark.unit
    def test_stats(self, gen_boron_removal_model):
        m = gen_boron_removal_model

        # Check to make sure we have the correct DOF and
        #   check to make sure the units are correct
        assert_units_consistent(m)
        assert degrees_of_freedom(m) == 10

        # set the variables
        model_setup(m, chem_add=1.8e-6)

        assert degrees_of_freedom(m) == 0

    @pytest.mark.unit
    def test_scaling(self, gen_boron_removal_model):
        m = gen_boron_removal_model

        # Set some scaling factors and look for 'bad' scaling
        scaling_setup(m)

        ## TODO: Revisit scaling for the generic package
        #   The unscaled vars come from the 'true-to-apparent'
        #   species vars in the generic package (which we don't
        #   ever use). May need to revisit how to scale
        # check that all variables have scaling factors
        unscaled_var_list = list(iscale.unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 8

    @pytest.mark.component
    def test_initialization(self, gen_boron_removal_model):
        m = gen_boron_removal_model
        initialization_tester(m)

        # check to make sure DOF does not change
        assert degrees_of_freedom(m) == 0

    @pytest.mark.component
    def test_solve(self, gen_boron_removal_model):
        m = gen_boron_removal_model

        # run solver and check for optimal solution
        results = solver.solve(m)
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, gen_boron_removal_model):
        m = gen_boron_removal_model

        assert value(
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", m.fs.unit.boron_name_id]
        ) == pytest.approx(1.20783e-4, rel=1e-4)
        assert value(
            m.fs.unit.outlet.flow_mol_phase_comp[0, "Liq", m.fs.unit.borate_name_id]
        ) == pytest.approx(8.02173e-5, rel=1e-4)
        assert value(m.fs.unit.outlet_pH()) == pytest.approx(9.0330, rel=1e-4)
        assert value(m.fs.unit.outlet_pOH()) == pytest.approx(4.9633, rel=1e-4)


# -----------------------------------------------------------------------------
# Start test class with bad config
class TestBoronRemoval_BadConfigs:
    @pytest.fixture(scope="class")
    def boron_removal_bad_configs(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # create dict to define ions (the prop pack requires this)
        ion_dict = {
            "solute_list": ["B[OH]3", "B[OH]4_-", "H_+", "OH_-", "Na_+"],
            "mw_data": {
                "H2O": 18e-3,
                "B[OH]3": 61.83e-3,
                "B[OH]4_-": 78.83e-3,
                "H_+": 1e-3,
                "OH_-": 17e-3,
                "Na_+": 23e-3,
            },
            "charge": {
                "B[OH]4_-": -1,
            },
        }

        # attach prop pack to flowsheet
        m.fs.properties = MCASParameterBlock(**ion_dict, ignore_neutral_charge=True)

        return m

    @pytest.mark.unit
    def test_build_failures(self, boron_removal_bad_configs):
        m = boron_removal_bad_configs

        # Invalid name of boron
        map = {
            "boron_name": "Boron",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "caustic_additive": {
                "cation_name": "Na_+",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = (
            "Given 'boron_name' {Boron} does not match any species "
            "name from the property package "
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Invalid name of borate
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "Borate",  # [is required]
            "caustic_additive": {
                "cation_name": "Na_+",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = (
            "Given 'borate_name' {Borate} does not match any species "
            "name from the property package "
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Empty dict
        map = {}
        error_msg = "Did not provide a 'dict' for 'chemical_mapping_data' "
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Invalid name of hydroxide
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "hydroxide_name": "Raccoon_City",
            "caustic_additive": {
                "cation_name": "Na_+",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = (
            "Given 'hydroxide_name' {Raccoon_City} does not match any species "
            "name from the property package "
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Invalid name of protons
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "proton_name": "Pulled_Pork",
            "caustic_additive": {
                "cation_name": "Na_+",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = (
            "Given 'proton_name' {Pulled_Pork} does not match any species "
            "name from the property package "
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Invalid name of cation
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "caustic_additive": {
                "cation_name": "Im_Batman",  # [is optional]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = (
            "Given 'cation_name' {Im_Batman} does not match any species "
            "name from the property package "
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Missing information (borate_name)
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "caustic_additive": {
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = "Missing some required information in 'chemical_mapping_data' "
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Missing information (mw_additive)
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "caustic_additive": {
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = "Missing some required information in 'chemical_mapping_data' "
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )

        # Improper data type (mw_additive must be a tuple with value and units)
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "caustic_additive": {
                "cation_name": "Na_+",  # [is required]
                "mw_additive": 40,  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = "Did not provide a tuple for 'mw_additive' "
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )


# -----------------------------------------------------------------------------
# Start test class with bad config
class TestBoronRemoval_BadConfigs_Generic:
    @pytest.fixture(scope="class")
    def boron_removal_bad_configs_gen(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # Configuration dictionary for generic
        thermo_config = {
            "components": {
                "H2O": {
                    "type": Solvent,
                    # Define the methods used to calculate the following properties
                    "dens_mol_liq_comp": Constant,
                    "enth_mol_liq_comp": Constant,
                    "cp_mol_liq_comp": Constant,
                    "entr_mol_liq_comp": Constant,
                    # Parameter data is always associated with the methods defined above
                    "parameter_data": {
                        "mw": (18.0153, pyunits.g / pyunits.mol),
                        "dens_mol_liq_comp_coeff": (
                            55.2,
                            pyunits.kmol * pyunits.m**-3,
                        ),
                        "cp_mol_liq_comp_coeff": (
                            75.312,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_liq_comp_ref": (0, pyunits.kJ / pyunits.mol),
                        "entr_mol_form_liq_comp_ref": (
                            0,
                            pyunits.J / pyunits.K / pyunits.mol,
                        ),
                    },
                    # End parameter_data
                },
                "H_+": {
                    "type": Cation,
                    "charge": 1,
                    # Define the methods used to calculate the following properties
                    "dens_mol_liq_comp": Constant,
                    "enth_mol_liq_comp": Constant,
                    "cp_mol_liq_comp": Constant,
                    "entr_mol_liq_comp": Constant,
                    # Parameter data is always associated with the methods defined above
                    "parameter_data": {
                        "mw": (1, pyunits.g / pyunits.mol),
                        "dens_mol_liq_comp_coeff": (
                            55.2,
                            pyunits.kmol * pyunits.m**-3,
                        ),
                        "cp_mol_liq_comp_coeff": (
                            75.312,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_liq_comp_ref": (0, pyunits.kJ / pyunits.mol),
                        "entr_mol_form_liq_comp_ref": (
                            0,
                            pyunits.J / pyunits.K / pyunits.mol,
                        ),
                    },
                    # End parameter_data
                },
                "Na_+": {
                    "type": Cation,
                    "charge": 1,
                    # Define the methods used to calculate the following properties
                    "dens_mol_liq_comp": Constant,
                    "enth_mol_liq_comp": Constant,
                    "cp_mol_liq_comp": Constant,
                    "entr_mol_liq_comp": Constant,
                    # Parameter data is always associated with the methods defined above
                    "parameter_data": {
                        "mw": (23, pyunits.g / pyunits.mol),
                        "dens_mol_liq_comp_coeff": (
                            55.2,
                            pyunits.kmol * pyunits.m**-3,
                        ),
                        "cp_mol_liq_comp_coeff": (
                            75.312,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_liq_comp_ref": (0, pyunits.kJ / pyunits.mol),
                        "entr_mol_form_liq_comp_ref": (
                            0,
                            pyunits.J / pyunits.K / pyunits.mol,
                        ),
                    },
                    # End parameter_data
                },
                "OH_-": {
                    "type": Anion,
                    "charge": -1,
                    # Define the methods used to calculate the following properties
                    "dens_mol_liq_comp": Constant,
                    "enth_mol_liq_comp": Constant,
                    "cp_mol_liq_comp": Constant,
                    "entr_mol_liq_comp": Constant,
                    # Parameter data is always associated with the methods defined above
                    "parameter_data": {
                        "mw": (17, pyunits.g / pyunits.mol),
                        "dens_mol_liq_comp_coeff": (
                            55.2,
                            pyunits.kmol * pyunits.m**-3,
                        ),
                        "cp_mol_liq_comp_coeff": (
                            75.312,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_liq_comp_ref": (0, pyunits.kJ / pyunits.mol),
                        "entr_mol_form_liq_comp_ref": (
                            0,
                            pyunits.J / pyunits.K / pyunits.mol,
                        ),
                    },
                    # End parameter_data
                },
                "B[OH]4_-": {
                    "type": Anion,
                    "charge": -1,
                    # Define the methods used to calculate the following properties
                    "dens_mol_liq_comp": Constant,
                    "enth_mol_liq_comp": Constant,
                    "cp_mol_liq_comp": Constant,
                    "entr_mol_liq_comp": Constant,
                    # Parameter data is always associated with the methods defined above
                    "parameter_data": {
                        "mw": (78.83, pyunits.g / pyunits.mol),
                        "dens_mol_liq_comp_coeff": (
                            55.2,
                            pyunits.kmol * pyunits.m**-3,
                        ),
                        "cp_mol_liq_comp_coeff": (
                            75.312,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_liq_comp_ref": (0, pyunits.kJ / pyunits.mol),
                        "entr_mol_form_liq_comp_ref": (
                            0,
                            pyunits.J / pyunits.K / pyunits.mol,
                        ),
                    },
                    # End parameter_data
                },
                "B[OH]3": {
                    "type": Solute,
                    "valid_phase_types": PT.aqueousPhase,
                    # Define the methods used to calculate the following properties
                    "dens_mol_liq_comp": Constant,
                    "enth_mol_liq_comp": Constant,
                    "cp_mol_liq_comp": Constant,
                    "entr_mol_liq_comp": Constant,
                    # Parameter data is always associated with the methods defined above
                    "parameter_data": {
                        "mw": (61.83, pyunits.g / pyunits.mol),
                        "dens_mol_liq_comp_coeff": (
                            55.2,
                            pyunits.kmol * pyunits.m**-3,
                        ),
                        "cp_mol_liq_comp_coeff": (
                            75.312,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),
                        "enth_mol_form_liq_comp_ref": (0, pyunits.kJ / pyunits.mol),
                        "entr_mol_form_liq_comp_ref": (
                            0,
                            pyunits.J / pyunits.K / pyunits.mol,
                        ),
                    },
                    # End parameter_data
                },
            },
            # End Component list
            "phases": {
                "Liq": {"type": AqueousPhase, "equation_of_state": Ideal},
            },
            "state_definition": FpcTP,
            "state_bounds": {
                "temperature": (273.15, 300, 650),
                "pressure": (5e4, 1e5, 1e6),
            },
            "pressure_ref": 1e5,
            "temperature_ref": 300,
            "base_units": {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            "inherent_reactions": {
                "H2O_Kw": {
                    "stoichiometry": {
                        ("Liq", "H2O"): -1,
                        ("Liq", "H_+"): 1,
                        ("Liq", "OH_-"): 1,
                    },
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_power_law_equil,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": (55.830, pyunits.J / pyunits.mol),
                        "k_eq_ref": (10**-14 / 55.2 / 55.2, pyunits.dimensionless),
                        "T_eq_ref": (298, pyunits.K),
                        # By default, reaction orders follow stoichiometry
                        #    manually set reaction order here to override
                        "reaction_order": {
                            ("Liq", "H2O"): 0,
                            ("Liq", "H_+"): 1,
                            ("Liq", "OH_-"): 1,
                        },
                    }
                    # End parameter_data
                }
                # End R1
            },
        }
        # End thermo_config definition

        # attach prop pack to flowsheet
        m.fs.properties = GenericParameterBlock(**thermo_config)

        return m

    @pytest.mark.unit
    def test_build_failures(self, boron_removal_bad_configs_gen):
        m = boron_removal_bad_configs_gen

        # Invalid name of boron
        map = {
            "boron_name": "B[OH]3",  # [is required]
            "borate_name": "B[OH]4_-",  # [is required]
            "caustic_additive": {
                "cation_name": "Na_+",  # [is required]
                "mw_additive": (40, pyunits.g / pyunits.mol),  # [is required]
                "moles_cation_per_additive": 1,  # [is required]
            },
        }
        error_msg = "Property Package CANNOT contain 'inherent_reactions' "
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.unit = BoronRemoval(
                property_package=m.fs.properties, chemical_mapping_data=map
            )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import value
from watertap.examples.flowsheets.full_treatment_train.model_components import (
    unit_separator,
    unit_0DRO,
    unit_1DRO,
    unit_ZONF,
)


@pytest.mark.component
def test_unit_separator_SepRO():
    m = unit_separator.solve_SepRO(base="TDS")
    assert value(
        m.fs.RO.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.4825, rel=1e-3)
    assert value(
        m.fs.RO.permeate.flow_mass_phase_comp[0, "Liq", "TDS"]
    ) == pytest.approx(3.5e-4, rel=1e-3)
    assert value(
        m.fs.RO.retentate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.4825, rel=1e-3)
    assert value(
        m.fs.RO.retentate.flow_mass_phase_comp[0, "Liq", "TDS"]
    ) == pytest.approx(3.465e-2, rel=1e-3)


@pytest.mark.component
def test_unit_separator_SepNF_ion():
    m = unit_separator.solve_SepNF(base="ion")
    assert value(
        m.fs.NF.mixed_state[0].mass_frac_phase_comp["Liq", "Na"]
    ) == pytest.approx(11122e-6, rel=1e-3)
    assert value(
        m.fs.NF.mixed_state[0].mass_frac_phase_comp["Liq", "Cl"]
    ) == pytest.approx(20317e-6, rel=1e-3)
    assert value(m.fs.NF.split_fraction[0, "permeate", "Cl"]) == pytest.approx(
        0.7753, rel=1e-3
    )
    assert value(
        m.fs.NF.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.8682, rel=1e-3)
    assert value(
        m.fs.NF.permeate.flow_mass_phase_comp[0, "Liq", "Na"]
    ) == pytest.approx(1.000e-2, rel=1e-3)
    assert value(
        m.fs.NF.permeate.flow_mass_phase_comp[0, "Liq", "Cl"]
    ) == pytest.approx(1.575e-2, rel=1e-3)
    assert value(
        m.fs.NF.retentate.flow_mass_phase_comp[0, "Liq", "Ca"]
    ) == pytest.approx(3.438e-4, rel=1e-3)
    assert value(
        m.fs.NF.retentate.flow_mass_phase_comp[0, "Liq", "Mg"]
    ) == pytest.approx(1.255e-3, rel=1e-3)
    assert value(
        m.fs.NF.retentate.flow_mass_phase_comp[0, "Liq", "SO4"]
    ) == pytest.approx(1.922e-3, rel=1e-3)


@pytest.mark.component
def test_unit_separator_SepNF_salt():
    m = unit_separator.solve_SepNF(base="salt")
    assert value(
        m.fs.NF.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.8682, rel=1e-3)
    assert value(
        m.fs.NF.permeate.flow_mass_phase_comp[0, "Liq", "NaCl"]
    ) == pytest.approx(2.544e-2, rel=1e-3)
    assert value(
        m.fs.NF.retentate.flow_mass_phase_comp[0, "Liq", "CaSO4"]
    ) == pytest.approx(1.168e-3, rel=1e-3)
    assert value(
        m.fs.NF.retentate.flow_mass_phase_comp[0, "Liq", "MgSO4"]
    ) == pytest.approx(1.376e-3, rel=1e-3)
    assert value(
        m.fs.NF.retentate.flow_mass_phase_comp[0, "Liq", "MgCl2"]
    ) == pytest.approx(3.401e-3, rel=1e-3)
    assert value(
        m.fs.NF.retentate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(9.645e-2, rel=1e-3)


@pytest.mark.component
def test_unit_0DRO_simple():
    m = unit_0DRO.solve_RO(base="TDS", level="simple")
    assert value(
        m.fs.RO.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.3389, rel=1e-3)
    assert value(
        m.fs.RO.permeate.flow_mass_phase_comp[0, "Liq", "TDS"]
    ) == pytest.approx(7.875e-5, rel=1e-3)
    assert value(
        m.fs.RO.retentate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.6261, rel=1e-3)
    assert value(
        m.fs.RO.retentate.flow_mass_phase_comp[0, "Liq", "TDS"]
    ) == pytest.approx(3.492e-2, rel=1e-3)


@pytest.mark.component
def test_unit_0DRO_detailed():
    m = unit_0DRO.solve_RO(base="TDS", level="detailed")
    assert value(
        m.fs.RO.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.2653, rel=1e-3)
    assert value(
        m.fs.RO.permeate.flow_mass_phase_comp[0, "Liq", "TDS"]
    ) == pytest.approx(8.476e-5, rel=1e-3)
    assert value(
        m.fs.RO.retentate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.6997, rel=1e-3)
    assert value(
        m.fs.RO.retentate.flow_mass_phase_comp[0, "Liq", "TDS"]
    ) == pytest.approx(3.492e-2, rel=1e-3)


@pytest.mark.component
def test_unit_1DRO_detailed():
    m = unit_1DRO.solve_RO(base="TDS", level="detailed")
    assert value(
        m.fs.RO.mixed_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
    ) == pytest.approx(0.2541, rel=1e-3)
    assert value(
        m.fs.RO.mixed_permeate[0].flow_mass_phase_comp["Liq", "TDS"]
    ) == pytest.approx(8.577e-5, rel=1e-3)
    assert value(
        m.fs.RO.retentate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.7109, rel=1e-3)
    assert value(
        m.fs.RO.retentate.flow_mass_phase_comp[0, "Liq", "TDS"]
    ) == pytest.approx(3.491e-2, rel=1e-3)


@pytest.mark.component
def test_unit_ZONF():
    m = unit_ZONF.solve_ZONF(base="ion")
    assert value(
        m.fs.NF.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.8350, rel=1e-3)
    assert value(
        m.fs.NF.permeate.flow_mass_phase_comp[0, "Liq", "Ca"]
    ) == pytest.approx(6.897e-5, rel=1e-3)
    assert value(
        m.fs.NF.permeate.flow_mass_phase_comp[0, "Liq", "Cl"]
    ) == pytest.approx(1.475e-2, rel=1e-3)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order feed block
"""
import pytest

from pyomo.environ import ConcreteModel, value, Var
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester

from watertap.unit_models.zero_order import FeedZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock

solver = get_solver()


class TestFeedZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(database=m.db)

        m.fs.unit = FeedZO(property_package=m.fs.params)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.fs.unit.outlet.flow_mass_comp, Var)
        for (t, j) in model.fs.unit.outlet.flow_mass_comp.keys():
            assert t == 0
            assert j in model.db.get_solute_set() or j == "H2O"

        assert isinstance(model.fs.unit.flow_vol, Var)
        assert len(model.fs.unit.flow_vol) == 1

        assert isinstance(model.fs.unit.conc_mass_comp, Var)
        for (t, j) in model.fs.unit.conc_mass_comp.keys():
            assert t == 0
            assert j in model.db.get_solute_set()

    @pytest.mark.unit
    def test_load_feed_data_from_database(self, model):
        data = model.db.get_source_data()

        model.fs.unit.load_feed_data_from_database()

        assert pytest.approx(data["default_flow"]["value"], rel=1e-12) == value(
            model.fs.unit.flow_vol[0]
        )
        assert model.fs.unit.flow_vol[0].fixed

        for j, v in data["solutes"].items():
            assert pytest.approx(v["value"], rel=1e-12) == value(
                model.fs.unit.conc_mass_comp[0, j]
            )
            assert model.fs.unit.conc_mass_comp[0, j].fixed

    @pytest.mark.unit
    def test_load_feed_data_from_database_no_overwrite(self, model):
        model.fs.unit.flow_vol[0].fix(42)
        model.fs.unit.conc_mass_comp[0, "tds"].fix(42)

        model.fs.unit.load_feed_data_from_database()

        # Should not have changed fixed variables
        assert 42 == value(model.fs.unit.flow_vol[0])
        assert model.fs.unit.flow_vol[0].fixed
        assert 42 == value(model.fs.unit.conc_mass_comp[0, "tds"])
        assert model.fs.unit.conc_mass_comp[0, "tds"].fixed

    @pytest.mark.unit
    def test_load_feed_data_from_database_overwrite(self, model):
        data = model.db.get_source_data()

        # Make sure overloaded variable still have set values
        assert 42 == value(model.fs.unit.flow_vol[0])
        assert model.fs.unit.flow_vol[0].fixed
        assert 42 == value(model.fs.unit.conc_mass_comp[0, "tds"])
        assert model.fs.unit.conc_mass_comp[0, "tds"].fixed

        model.fs.unit.load_feed_data_from_database(overwrite=True)

        # This should have reset the value of all variables
        assert pytest.approx(data["default_flow"]["value"], rel=1e-12) == value(
            model.fs.unit.flow_vol[0]
        )
        assert model.fs.unit.flow_vol[0].fixed

        for j, v in data["solutes"].items():
            assert pytest.approx(v["value"], rel=1e-12) == value(
                model.fs.unit.conc_mass_comp[0, j]
            )
            assert model.fs.unit.conc_mass_comp[0, j].fixed

    @pytest.mark.unit
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model)

    @pytest.mark.unit
    def test_load_feed_data_from_database_missing(self, model, caplog):
        data = model.db.get_source_data()

        del data["default_flow"]
        del data["solutes"]["tds"]

        model.fs.unit.load_feed_data_from_database(overwrite=True)

        assert (
            "fs.unit no default flowrate was defined in database water "
            "source. Value was not fixed."
        ) in caplog.text
        assert (
            "fs.unit component tds was not defined in database water "
            "source. Value was not fixed."
        ) in caplog.text


@pytest.mark.component
def test_initialize():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.params = WaterParameterBlock(database=m.db)

    m.fs.unit = FeedZO(property_package=m.fs.params)

    m.fs.unit.load_feed_data_from_database()

    initialization_tester(m)

    res = {
        "H2O": 4263.816948529999,
        "boron": 0.020166519999999997,
        "bromide": 0.3038727899999999,
        "calcium": 1.8773196799999994,
        "chloride": 87.82519459999996,
        "magnesium": 5.857457399999998,
        "potassium": 1.8117784899999994,
        "sodium": 48.945060699999985,
        "strontium": 0.005958289999999999,
        "sulfate": 12.283243999999996,
        "tds": 160.41549999999995,
        "tss": 0.13749899999999993,
    }

    for (t, j), v in m.fs.unit.outlet.flow_mass_comp.items():
        assert value(v) == pytest.approx(res[j], rel=1e-5)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from abc import abstractmethod, ABC


class ParallelManager(ABC):
    ROOT_PROCESS_RANK = 0

    @abstractmethod
    def is_root_process(self):
        """
        Return whether the current process should be considered as the root of the parallel group it
        is a part of.
        """
        raise NotImplementedError

    @abstractmethod
    def get_rank(self):
        """
        Get the process's rank number in its parallel peer group.
        """
        raise NotImplementedError

    @abstractmethod
    def number_of_worker_processes(self):
        """
        Return how many total processes are running local simulations.
        """
        raise NotImplementedError

    @abstractmethod
    def sync_with_peers(self):
        """
        Implement a synchronization point with any peer processes.
        """
        raise NotImplementedError

    @abstractmethod
    def sync_array_with_peers(self, data):
        """
        Synchronize an array with all peers processes. The data parameter is either:
        - (if root) the source of truth that will be broadcast to all processes
        - (if not root) an empty buffer that will hold the data sent from root once the function returns
        """
        raise NotImplementedError

    @abstractmethod
    def sync_pyobject_with_peers(self, obj):
        """
        Synchronize a python object with all peer processes. The obj parameter is either:
        - (if root) the source of truth that will be broadcast to all processes
        - (if not root) ignored

        Different from sync_array_with_peers in that it returns the synced object rather than
        using an existing buffer.
        """
        raise NotImplementedError

    @abstractmethod
    def combine_data_with_peers(self, data):
        """
        Combine the data from each peer into a list. The return value will be a list of all
        the data elements provided by each process that calls this function.

        With multiple processes, this must:
        - act as a synchronization point
        - return the list in the same order on each process
        """
        raise NotImplementedError

    @abstractmethod
    def gather_arrays_to_root(self, sendbuf, recvbuf_spec):
        """
        Gather the data in the send buffer to the root process. Parameters are:
        - sendbuf: the data to be sent from this process
        - recvbuf_spec: a tuple of (receive buffer, list of sizes) where the list of sizes is
        how much data will come from each process. Ignored if not the root process.
        """
        raise NotImplementedError

    @abstractmethod
    def sum_values_and_sync(self, sendbuf, recvbuf):
        """
        Sum a list of values from each process and sync the result to all processes. Parameters:
        - sendbuf: an array of values the local process is contributing to the sum
        - recvbuf: an array of a single value that will hold the summed result (same on
        each process) when the function returns.

        Note: this is a specific case of a global reduce (with sum() as the reducing function).
        If more than sum() is needed in the future, this function should be extended to receive
        an Op argument.
        """
        raise NotImplementedError

    @abstractmethod
    def scatter(
        self,
        do_build,
        do_build_kwargs,
        do_execute,
        all_parameters,
    ):
        """
        Scatter the specified execution out, as defined by the implementation's parallelism, for
        a list of parameters.
        Args:
        - do_build: a function that builds the arguments necessary for the execution function. expected to
        return a list that will be exploded and passed into the do_execute function as arguments.
        - do_build_kwargs: a dictionary of keyword arguments for the do_build function
        - do_execute: the execution function. expected to take in the list of local parameters
        as its first argument. any arguments after that should match the list returned by the
        do_build function.
        - all_parameters: a list of all parameters to be run. included so that
        different implementations of the parallel manager can make decisions about splitting and
        parallelization.
        """
        raise NotImplementedError

    @abstractmethod
    def gather(self):
        """
        Gather the results of the computation that was kicked off via a previous scatter.
        Returns:
        - a list of LocalResults, representing the results for each process. each result will
        be the return value of the do_execute function from scatter() for one process.
        """
        raise NotImplementedError

    @abstractmethod
    def results_from_local_tree(self, results):
        """
        Given a list of LocalResults objects, return a sublist of the ones the current process
        is responsible for.
        """
        raise NotImplementedError


# TODO this probably should be owned by parameer sweep as its PS specific


def build_and_execute(do_build, do_build_kwargs, do_execute, local_parameters):
    """
    Entrypoint for implementations of the parallel manager to use for running the
    build and execute functions. Defined at the top level so that it's picklable.

    For a description of the first three arguments, see the scatter() function above.
    The fourth argument is the list of local parameters that should be run by this process.
    """
    pa = parallelActor(do_build, do_build_kwargs, do_execute, local_parameters)
    results = pa.execute(local_parameters)
    # execute_args = do_build(**do_build_kwargs)
    # results = do_execute(local_parameters, *execute_args)
    return results


# TODO this probably should be owned by parameer sweep as its PS specific
class parallelActor:
    def __init__(self, do_build, do_build_kwargs, do_execute, local_parameters):
        self.do_build = do_build
        self.do_build_kwargs = do_build_kwargs
        self.do_execute = do_execute
        self.local_parameters = local_parameters
        self.build_model()

    def build_model(self):
        self.exec_args = self.do_build(**self.do_build_kwargs)

    def execute(self, local_parameters):
        result = self.do_execute(local_parameters, *self.exec_args)
        return result



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo

from pyomo.util.calc_var_value import calculate_variable_from_constraint

from idaes.core import declare_process_block_class
from idaes.core.base.costing_base import register_idaes_currency_units

from watertap.costing.costing_base import WaterTAPCostingBlockData


@declare_process_block_class("WaterTAPCosting")
class WaterTAPCostingData(WaterTAPCostingBlockData):
    def build_global_params(self):
        # Register currency and conversion rates based on CE Index
        register_idaes_currency_units()

        self._build_common_global_params()

        # Set the base year for all costs
        self.base_currency = pyo.units.USD_2018
        # Set a base period for all operating costs
        self.base_period = pyo.units.year

        # Build flowsheet level costing components
        # These are the global parameters
        self.factor_total_investment = pyo.Var(
            initialize=1.0,
            doc="Total investment factor [investment cost/equipment cost]",
            units=pyo.units.dimensionless,
        )
        self.factor_maintenance_labor_chemical = pyo.Var(
            initialize=0.03,
            doc="Maintenance-labor-chemical factor [fraction of equipment cost/year]",
            units=pyo.units.year**-1,
        )
        self.factor_capital_annualization = pyo.Var(
            initialize=0.1,
            doc="Capital annualization factor [fraction of investment cost/year]",
            units=pyo.units.year**-1,
        )
        self.capital_recovery_factor.expr = self.factor_capital_annualization

        # fix the parameters
        self.fix_all_vars()

    def build_process_costs(self):
        # add total_captial_cost and total_operating_cost
        self._build_common_process_costs()

        self.total_capital_cost_constraint = pyo.Constraint(
            expr=self.total_capital_cost
            == self.factor_total_investment * self.aggregate_capital_cost
        )

        self.maintenance_labor_chemical_operating_cost = pyo.Expression(
            expr=self.factor_maintenance_labor_chemical * self.aggregate_capital_cost,
            doc="Maintenance-labor-chemical operating cost",
        )

        self.total_fixed_operating_cost = pyo.Expression(
            expr=self.aggregate_fixed_operating_cost
            + self.maintenance_labor_chemical_operating_cost,
            doc="Total fixed operating costs",
        )

        self.total_variable_operating_cost = pyo.Expression(
            expr=(
                self.aggregate_variable_operating_cost
                + sum(self.aggregate_flow_costs[f] for f in self.used_flows)
                * self.utilization_factor
            )
            if self.used_flows
            else self.aggregate_variable_operating_cost,
            doc="Total variable operating cost of process per operating period",
        )

        self.total_operating_cost_constraint = pyo.Constraint(
            expr=self.total_operating_cost
            == (self.total_fixed_operating_cost + self.total_variable_operating_cost),
            doc="Total operating cost of process per operating period",
        )

        self.total_annualized_cost = pyo.Expression(
            expr=(
                self.total_capital_cost * self.capital_recovery_factor
                + self.total_operating_cost
            ),
            doc="Total annualized cost of operation",
        )

    def initialize_build(self):
        calculate_variable_from_constraint(
            self.total_capital_cost, self.total_capital_cost_constraint
        )
        calculate_variable_from_constraint(
            self.total_operating_cost, self.total_operating_cost_constraint
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a landfill unit.
"""

from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_pt, constant_intensity, ZeroOrderBaseData

# Some more inforation about this module
__author__ = "Chenyu Wang"


@declare_process_block_class("LandfillZO")
class LandfillZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a landfill unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "landfill"

        build_pt(self)
        constant_intensity(self)

        self.capacity_basis = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hr,
            doc="capacity basis for capital cost",
        )

        self.total_mass = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hr,
            doc="total mass flow rate",
        )

        self._fixed_perf_vars.append(self.capacity_basis)

        @self.Constraint(self.flowsheet().time, doc="Total mass constraint")
        def total_mass_constraint(b, t):
            return b.total_mass[t] == sum(
                pyunits.convert(
                    b.inlet.flow_mass_comp[t, m], to_units=pyunits.kg / pyunits.hr
                )
                for m in b.config.property_package.component_list
            )

        self._perf_var_dict["Capacity Basis (kg/hr)"] = self.capacity_basis
        self._perf_var_dict["Total Mass (kg/hr)"] = self.total_mass

    @property
    def default_costing_method(self):
        return self.cost_landfill

    @staticmethod
    def cost_landfill(blk, number_of_parallel_units=1):
        """
        General method for costing landfill. Capital cost is based on the total mass and
        capacity basis.
        Args:
            number_of_parallel_units (int, optional) - cost this unit as
                        number_of_parallel_units parallel units (default: 1)
        """

        t0 = blk.flowsheet().time.first()
        sizing_term = blk.unit_model.total_mass[t0] / blk.unit_model.capacity_basis[t0]

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        A, B = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter", "capital_b_parameter"],
        )

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Call general power law costing method
        blk.unit_model._general_power_law_form(
            blk, A, B, sizing_term, factor, number_of_parallel_units
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
    This file demonstrates how to use EDB to create a simple acid problem.

    (1) Before we can start, you must install MongoDB (which is installed separately)

        [See more information on the ReadTheDocs under 'Getting Started --> Installing WaterTAP']


    (2) After installing MongoDB, you will need to 'load' the database using the
        command line function 'edb load -b'. This will load the default database
        that WaterTAP is bundled with.

        [NOTE: If you need to 'reload' the database, simply use the command 'edb drop -d electrolytedb'
                in the command line. The database on MongoDB is named "electrolytedb"]

        [NOTE 2: You can invoke the command line utility with the "help" keyword to
                get more information on funtionality. Command: 'edb --help' or 'edb [arg] --help']


    (3) To use EDB in python, start by importing the interface class object 'ElectrolyteDB'


    (4) Invoke the 'ElectrolyteDB' object to connect to the database


    (5) Grab a 'base' for a configuration dictionary, and place it into a class object
        This time, we will grab a base that is for a Liq only problem using FpcTP
        state variables.


    (6) Get the chemcial species/components for a simulation case. There are a number of ways
        to do this. In this example, we will grab them by finding all components that contain
        only specific elements. Then, we add those components and their associated parameters
        to the configuration dictionary being built from the 'base'.

        [NOTE: An alternative method is to provide a list of the names of components you want]


    (7) Get the set of reactions you want in your system and put into a 'base' object.
        In this case, we are getting all reactions associated with a system of water
        and carbonic acid. We should get three reactions::

            H2O <--> H_+ + OH_-
            H2CO3 <--> H_+ + HCO3_-
            HCO3_- <--> H_+ + CO3_2-


    (8) When using an reactor object in IDAES, you must always provide a 'reaction_config'
        to match with the 'thermo_config'. We can create a base 'reaction' config from
        the database and add reactions to that config in the same way we do for the
        'thermo_config' when adding reactions as inherent.

        [NOTE: If a reaction is added to a 'thermo_config' as 'inherent', it should
               NOT be added to a 'reaction_config' as 'equilibrium']


    (9) [NEW Step] Build an equilibrium reactor from the 'thermo_config' and 'reaction_config'
        that were generated from the EDB.

"""

# ========= These imports (below) are for testing the configs from EDB ===============
# Import specific pyomo objects
from pyomo.environ import (
    ConcreteModel,
)

# Import the idaes objects for Generic Properties and Reactions
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.base.generic_reaction import (
    GenericReactionParameterBlock,
)

# Import the idaes object for the EquilibriumReactor unit model
from idaes.models.unit_models.equilibrium_reactor import EquilibriumReactor

# Import the core idaes objects for Flowsheets and types of balances
from idaes.core import FlowsheetBlock

# ========= These imports (above) are for testing the configs from EDB ===============


# ========================== (3 & 4) ================================
# Import ElectrolyteDB object
from watertap.examples.edb.the_basics import grab_base_reaction_config

__author__ = "Austin Ladshaw"


# ========================== (5) ================================
# Grab a new base config for our thermo, but this time we will use
#   one of the newer bases that will use the FpcTP state vars and
#   a Liq only system.
def grab_thermo_Liq_FpcTP_base(db):
    # Get the base and place into a result object
    base = db.get_base("thermo_Liq_FpcTP")
    return base


# ========================== (6) ================================
# Get chemical components/species for a simulation case
#       NOTE: This function here also returns a 'list' of the
#           components that it finds. This is not a built in
#           feature of the EDB, but is very useful because
#           getting reactions is dependent on the component list.
def get_components_and_add_to_idaes_config(db, base_obj, comp_list):
    res_obj_comps = db.get_components(component_names=comp_list)

    # Iterate through the results object and add the components
    #   to the base_obj
    for comp_obj in res_obj_comps:
        print("Adding " + str(comp_obj.name) + "")
        base_obj.add(comp_obj)
    print()
    return base_obj


# ========================== (7) ================================
# Grab the reactions associated with the list of components and add
#   them to a reaction base as equilibrium reactions
#
def add_equilibrium_reactions_to_react_base(db, react_base_obj, comp_list):
    react_obj = db.get_reactions(component_names=comp_list)
    for r in react_obj:
        print("Found reaction: " + str(r.name))
        react_base_obj.add(r)
    return react_base_obj


# ========================== (9) ================================
# Create the Pyomo model by using the thermo_config and reaction_config
#   that were generated from the EDB.
#
def build_equilibrium_model(thermo_config, reaction_config):
    model = ConcreteModel()
    model.fs = FlowsheetBlock(dynamic=False)
    model.fs.thermo_params = GenericParameterBlock(**thermo_config)
    model.fs.rxn_params = GenericReactionParameterBlock(
        property_package=model.fs.thermo_params, **reaction_config
    )

    model.fs.unit = EquilibriumReactor(
        property_package=model.fs.thermo_params,
        reaction_package=model.fs.rxn_params,
        has_rate_reactions=False,
        has_equilibrium_reactions=True,
        has_heat_transfer=False,
        has_heat_of_reaction=False,
        has_pressure_change=False,
    )

    return model


# Run script for testing
def run_simple_acid_with_mockdb(db):
    base_obj = grab_thermo_Liq_FpcTP_base(db)

    # Our components for this problem are as follows:
    comp_list = ["H2O", "H_+", "OH_-", "H2CO3", "HCO3_-", "CO3_2-"]

    base_obj = get_components_and_add_to_idaes_config(db, base_obj, comp_list)

    # Create a reaction config
    react_base = grab_base_reaction_config(db)

    # Add reactions to the reaction base as 'equilibrium'
    react_base = add_equilibrium_reactions_to_react_base(db, react_base, comp_list)

    # Now, we can actually see if we created a correct model by looking
    #       for degrees of freedom, state variables, etc.
    thermo_config = base_obj.idaes_config
    reaction_config = react_base.idaes_config
    model = build_equilibrium_model(thermo_config, reaction_config)

    return model



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from ..util import (
    register_costing_parameter_block,
    cost_rectifier,
    make_capital_cost_var,
    make_fixed_operating_cost_var,
)


def build_electrodialysis_cost_param_block(blk):
    # The following costing itemization and values are referenced to "Desalination 452 (2019) 265–278"
    blk.membrane_capital_cost = pyo.Var(
        initialize=160,
        doc="Membrane and capitcal costs in [US$/m^2-membrane-area]",
        units=pyo.units.USD_2018 / (pyo.units.meter**2),
    )

    blk.factor_membrane_replacement = pyo.Var(
        initialize=0.2,
        doc="Membrane and equipment (other stack components) housing replacement factor, equal to 1/lifetime.",
        units=pyo.units.year**-1,
    )

    blk.stack_electrode_captical_cost = pyo.Var(
        initialize=2100,
        doc="Electrode cost in [US$/m^2-electrode-area] ",
        units=pyo.units.USD_2018 / (pyo.units.meter**2),
    )

    blk.factor_stack_electrode_replacement = pyo.Var(
        initialize=0.2,
        doc="Stack and electrode replacement factor, equal to 1/lifetime.",
        units=pyo.units.year**-1,
    )


@register_costing_parameter_block(
    build_rule=build_electrodialysis_cost_param_block,
    parameter_block_name="electrodialysis",
)
def cost_electrodialysis(blk, cost_electricity_flow=True, has_rectifier=False):
    """
    Function for costing the Electrodialysis unit

    Args:
        cost_electricity_flow (:obj:`bool`, optional): Option for including the
            costing of electricity. Defaults to True.
        has_rectifier (:obj:`bool`, optional): Option for including a rectifier.
            Defaults to False.
    """
    t0 = blk.flowsheet().time.first()

    # Changed this to grab power from performance table which is identified
    # by same key regardless of whether the Electrodialysis unit is 0D or 1D
    if cost_electricity_flow:
        if not has_rectifier:
            blk.costing_package.cost_flow(
                pyo.units.convert(
                    blk.unit_model.get_power_electrical(t0),
                    to_units=pyo.units.kW,
                ),
                "electricity",
            )
        else:
            power = blk.unit_model.get_power_electrical(blk.flowsheet().time.first())
            cost_rectifier(blk, power=power, ac_dc_conversion_efficiency=0.9)
    cost_electrodialysis_stack(blk)


def cost_electrodialysis_stack(blk):
    """
    Generic function for costing the stack in an electrodialysis unit.
    Assumes the unit_model has a `cell_pair_num`, `cell_width`, and `cell_length`
    set of variables used to size the total membrane area.

    """
    make_capital_cost_var(blk)
    make_fixed_operating_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    if blk.find_component("capital_cost_rectifier") is not None:
        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost
            == blk.cost_factor
            * (
                pyo.units.convert(
                    blk.costing_package.electrodialysis.membrane_capital_cost
                    * (
                        2
                        * blk.unit_model.cell_pair_num
                        * blk.unit_model.cell_width
                        * blk.unit_model.cell_length
                    )
                    + blk.costing_package.electrodialysis.stack_electrode_captical_cost
                    * (2 * blk.unit_model.cell_width * blk.unit_model.cell_length),
                    to_units=blk.costing_package.base_currency,
                )
                + blk.capital_cost_rectifier
            )
        )
    else:
        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost
            == blk.cost_factor
            * pyo.units.convert(
                blk.costing_package.electrodialysis.membrane_capital_cost
                * (
                    2
                    * blk.unit_model.cell_pair_num
                    * blk.unit_model.cell_width
                    * blk.unit_model.cell_length
                )
                + blk.costing_package.electrodialysis.stack_electrode_captical_cost
                * (2 * blk.unit_model.cell_width * blk.unit_model.cell_length),
                to_units=blk.costing_package.base_currency,
            )
        )
    blk.fixed_operating_cost_constraint = pyo.Constraint(
        expr=blk.fixed_operating_cost
        == pyo.units.convert(
            blk.costing_package.electrodialysis.factor_membrane_replacement
            * blk.costing_package.electrodialysis.membrane_capital_cost
            * (
                2
                * blk.unit_model.cell_pair_num
                * blk.unit_model.cell_width
                * blk.unit_model.cell_length
            )
            + blk.costing_package.electrodialysis.factor_stack_electrode_replacement
            * blk.costing_package.electrodialysis.stack_electrode_captical_cost
            * (2 * blk.unit_model.cell_width * blk.unit_model.cell_length),
            to_units=blk.costing_package.base_currency
            / blk.costing_package.base_period,
        )
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
    This file demonstrates the basics of working with and using the electrolyte
    database (EDB).

    (1) Before we can start, you must install MongoDB (which is installed separately)

        [See more information on the ReadTheDocs under 'Getting Started --> Installing WaterTAP']


    (2) After installing MongoDB, you will need to 'load' the database using the
        command line function 'edb load -b'. This will load the default database
        that WaterTAP is bundled with.

        [NOTE: If you need to 'reload' the database, simply use the command 'edb drop -d electrolytedb'
                in the command line. The database on MongoDB is named "electrolytedb"]

        [NOTE 2: You can invoke the command line utility with the "help" keyword to
                get more information on funtionality. Command: 'edb --help' or 'edb [arg] --help']


    (3) To use EDB in python, start by importing the interface class object 'ElectrolyteDB'


    (4) Invoke the 'ElectrolyteDB' object to connect to the database


    (5) Grab a 'base' for a configuration dictionary, and place it into a class object


    (6) Get the chemcial species/components for a simulation case. There are a number of ways
        to do this. In this example, we will grab them by finding all components that contain
        only specific elements. Then, we add those components and their associated parameters
        to the configuration dictionary being built from the 'base'.

        [NOTE: An alternative method is to provide a list of the names of components you want]


    (7) Get the set of reactions you want in your system and put into a 'base' object.
        That 'base' can be either a 'thermo' base or a 'reaction' (as in this case)
        base. IF you are adding reactions to a 'thermo' base, they should be added
        as 'inherent' reactions. IF you are adding reactions to a 'reaction' base,
        they should be added as 'equilibrium' (or other) reactions.


    (8) When using an reactor object in IDAES, you must always provide a 'reaction_config'
        to match with the 'thermo_config'. We can create a base 'reaction' config from
        the database and add reactions to that config in the same way we do for the
        'thermo_config' when adding reactions as inherent.

        [NOTE: If a reaction is added to a 'thermo_config' as 'inherent', this it should
               NOT be added to a 'reaction_config' as 'equilibrium']

"""

# ========= These imports (below) are for testing the configs from EDB ===============
# Import specific pyomo objects
from pyomo.environ import (
    ConcreteModel,
)

# Import the idaes objects for Generic Properties and Reactions
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.base.generic_reaction import (
    GenericReactionParameterBlock,
)

# Import the idaes object for the EquilibriumReactor unit model
from idaes.models.unit_models.equilibrium_reactor import EquilibriumReactor

# Import the core idaes objects for Flowsheets and types of balances
from idaes.core import FlowsheetBlock

# ========= These imports (above) are for testing the configs from EDB ===============


# ========================== (3) ================================
# Import ElectrolyteDB object
from watertap.edb import ElectrolyteDB

__author__ = "Austin Ladshaw"

# ========================== (4) ================================
# By default, invoking the 'ElectrolyteDB' object (with no args)
#   will attempt to connect to the local host database. You can
#   check the connection by calling the 'can_connect' function
#   and passing the 'host' and 'port' as args. If no 'host' or
#   'port' are given, then it uses the defaults.
def connect_to_edb(test_invalid_host=False):
    print("connecting to " + str(ElectrolyteDB.DEFAULT_URL))
    db = ElectrolyteDB()
    connected = db.can_connect()
    return (db, connected)


# ========================== (5) ================================
# All configuration files used in WaterTAP for electrolyte chemistry
#   require a 'base' dictionary to start. For example, we need to
#   create a 'thermo_config' dictionary to pass to the GenericProperty
#   package in IDAES. That 'thermo_config' file will always have a
#   few specific items in common with most other configuration files.
#   Thus, this operation will populate a class object that holds the
#   data assocated with that 'base' dictionary.
#
# In the EDB, there are several different 'base' structures to start
#   from. In this example, we will build from the 'default_thermo'
#   configuration base.
def grab_base_thermo_config(db):
    # Get the base and place into a result object
    base = db.get_base("default_thermo")
    return base


# ========================== (6) ================================
# Get chemical components/species for a simulation case
#       NOTE: This function here also returns a 'list' of the
#           components that it finds. This is not a built in
#           feature of the EDB, but is very useful because
#           getting reactions is dependent on the component list.
def get_components_and_add_to_idaes_config(db, base_obj, by_elements=False):
    # Going to grab all components that contain ONLY "H" and "O"
    #   Expected behavior = Will get "H2O", "H_+", and "OH_-"
    element_list = ["H", "O"]

    # Alternatively, you can pass a list of individual componets
    #   you want to grab and the EDB functions should grab explicitly
    #   those components/species you want.
    comp_list = ["H2O", "H_+", "OH_-"]

    # Just like before, this function returns a results object
    #   that contains other objects that must be iterated through
    #   in order to access the information. Then, call the 'add'
    #   function to add those components to the 'base' object
    if by_elements == True:
        res_obj_comps = db.get_components(element_names=element_list)
    else:
        res_obj_comps = db.get_components(component_names=comp_list)

    # Iterate through the results object and add the components
    #   to the base_obj
    db_comp_list = []
    for comp_obj in res_obj_comps:
        print("Adding " + str(comp_obj.name) + "")
        base_obj.add(comp_obj)
        db_comp_list.append(comp_obj.name)
    print()
    return (base_obj, db_comp_list)


# ========================== (7) ================================
# Grab the reactions associated with the list of components and add
#   them to a base object (which could be a 'thermo' base or 'reaction' base)
#
def get_reactions_return_object(db, base_obj, comp_list, is_inherent=True):
    react_obj = db.get_reactions(component_names=comp_list)
    for r in react_obj:
        print("Found reaction: " + str(r.name))
        if is_inherent == True:
            r._data["type"] = "inherent"
        base_obj.add(r)
    return base_obj


# ========================== (8) ================================
# Create a base config for reactions.
def grab_base_reaction_config(db):
    # Get the base and place into a result object
    base = db.get_base("reaction")
    return base


# This function will produce an error if the thermo config is not correct
def is_thermo_config_valid(thermo_config):
    model = ConcreteModel()
    model.fs = FlowsheetBlock(dynamic=False)
    model.fs.thermo_params = GenericParameterBlock(**thermo_config)
    return True


# This function will produce an error if the thermo config is not correct
#   or if the pairing of the thermo and reaction config are invalid
def is_thermo_reaction_pair_valid(thermo_config, reaction_config):
    model = ConcreteModel()
    model.fs = FlowsheetBlock(dynamic=False)
    model.fs.thermo_params = GenericParameterBlock(**thermo_config)
    model.fs.rxn_params = GenericReactionParameterBlock(
        property_package=model.fs.thermo_params, **reaction_config
    )
    return True


# Run script for testing
def run_the_basics_with_mockdb(db):
    base_obj = grab_base_thermo_config(db)

    (base_obj, comp_list) = get_components_and_add_to_idaes_config(db, base_obj)

    # Create a reaction config
    react_base = grab_base_reaction_config(db)

    # Add reactions to the reaction base as 'equilibrium'
    react_base = get_reactions_return_object(
        db, react_base, comp_list, is_inherent=False
    )

    # If all goes well, this function returns true
    return is_thermo_reaction_pair_valid(base_obj.idaes_config, react_base.idaes_config)


# Run script for testing
def run_the_basics_alt_with_mockdb(db):
    base_obj = grab_base_thermo_config(db)

    (base_obj, comp_list) = get_components_and_add_to_idaes_config(
        db, base_obj, by_elements=True
    )

    # Add reactions to the thermo base as 'inherent'
    base_obj = get_reactions_return_object(db, base_obj, comp_list, is_inherent=True)

    # If all goes well, this function returns true
    return is_thermo_config_valid(base_obj.idaes_config)


# Run script for testing
def run_the_basics_dummy_rxn_with_mockdb(db):
    base_obj = grab_base_thermo_config(db)

    (base_obj, comp_list) = get_components_and_add_to_idaes_config(
        db, base_obj, by_elements=True
    )

    # Add reactions to the thermo base as 'inherent'
    base_obj = get_reactions_return_object(db, base_obj, comp_list, is_inherent=True)

    # Create a reaction config
    react_base = grab_base_reaction_config(db)

    # If no reactions are in the reaction base, this will cause an error in IDAES.
    #   However, we can add a 'dummy' reaction just to satisfy the IDAES code base.
    react_obj = db.get_reactions(reaction_names=["dummy"])
    for r in react_obj:
        print("Found reaction: " + str(r.name))
        react_base.add(r)

    # IDAES will throw an exception when we try to do this if something is wrong
    thermo_config = base_obj.idaes_config
    reaction_config = react_base.idaes_config
    model = ConcreteModel()
    model.fs = FlowsheetBlock(dynamic=False)
    model.fs.thermo_params = GenericParameterBlock(**thermo_config)
    model.fs.rxn_params = GenericReactionParameterBlock(
        property_package=model.fs.thermo_params, **reaction_config
    )

    model.fs.unit = EquilibriumReactor(
        property_package=model.fs.thermo_params,
        reaction_package=model.fs.rxn_params,
        has_rate_reactions=False,
        has_equilibrium_reactions=False,
        has_heat_transfer=False,
        has_heat_of_reaction=False,
        has_pressure_change=False,
    )

    # If all goes well, this function returns true
    return is_thermo_reaction_pair_valid(base_obj.idaes_config, react_base.idaes_config)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from pyomo.environ import (
    ConcreteModel,
    Expression,
    Constraint,
    Param,
    TransformationFactory,
)
from pyomo.network import Arc
from idaes.core import FlowsheetBlock
from idaes.core.util.scaling import (
    calculate_scaling_factors,
)
from idaes.core.util.initialization import propagate_state
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import (
    pretreatment_NF,
    gypsum_saturation_index,
    translator_block,
    costing,
)
from watertap.examples.flowsheets.full_treatment_train.model_components import (
    property_models,
)
from watertap.examples.flowsheets.full_treatment_train.util import (
    solve_block,
    check_dof,
)


def build_components(m, has_bypass=True):
    # build flowsheet
    property_models.build_prop(m, base="ion")
    pretrt_port = pretreatment_NF.build_pretreatment_NF(
        m, NF_type="ZO", NF_base="ion", has_bypass=has_bypass
    )

    property_models.build_prop(m, base="TDS")
    translator_block.build_tb(
        m, base_inlet="ion", base_outlet="TDS", name_str="tb_pretrt_to_desal"
    )

    # Arc to translator block
    m.fs.s_pretrt_tb = Arc(
        source=pretrt_port["out"], destination=m.fs.tb_pretrt_to_desal.inlet
    )

    property_models.build_prop(m, base="eNRTL")
    gypsum_saturation_index.build(m, section="pretreatment")

    m.fs.NF.area.fix(175)
    if has_bypass:
        m.fs.splitter.split_fraction[0, "bypass"].fix(0.50)

        m.fs.removal_Ca = Expression(
            expr=(
                m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "Ca"]
                - m.fs.mixer.mixed_state[0].flow_mass_phase_comp["Liq", "Ca"]
            )
            / m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "Ca"]
        )
        m.fs.removal_Mg = Expression(
            expr=(
                m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "Mg"]
                - m.fs.mixer.mixed_state[0].flow_mass_phase_comp["Liq", "Mg"]
            )
            / m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "Mg"]
        )


def build(m, has_bypass=True):
    """
    Build a flowsheet with nanofiltration as the pretreatment process.
    """
    build_components(m, has_bypass=has_bypass)

    # annual water production
    m.fs.treated_flow_vol = Expression(
        expr=m.fs.tb_pretrt_to_desal.properties_out[0].flow_vol
    )
    costing.build_costing(m, NF_type="ZO")

    return m


def scale(m, has_bypass=True):
    pretreatment_NF.scale_pretreatment_NF(
        m, NF_type="ZO", NF_base="ion", has_bypass=has_bypass
    )
    calculate_scaling_factors(m.fs.tb_pretrt_to_desal)


def initialize(m, has_bypass=True):
    optarg = {"nlp_scaling_method": "user-scaling"}
    pretreatment_NF.initialize_pretreatment_NF(
        m, NF_type="ZO", NF_base="ion", has_bypass=has_bypass
    )
    m.fs.pretrt_saturation.properties.initialize(optarg=optarg)
    propagate_state(m.fs.s_pretrt_tb)
    m.fs.tb_pretrt_to_desal.initialize(optarg=optarg)


def report(m, has_bypass=True):
    pretreatment_NF.display_pretreatment_NF(
        m, NF_type="ZO", NF_base="ion", has_bypass=has_bypass
    )
    m.fs.tb_pretrt_to_desal.report()


def solve_flowsheet(has_bypass=True):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    build(m, has_bypass=has_bypass)
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scale
    scale(m, has_bypass=has_bypass)
    calculate_scaling_factors(m)

    # initialize
    initialize(m, has_bypass=has_bypass)

    check_dof(m)
    solve_block(m, tee=True, fail_flag=True)

    # report
    report(m, has_bypass=has_bypass)

    return m


def simulate(m, check_termination=True):
    return solve_block(m, tee=False, fail_flag=check_termination)


def set_optimization_components(m, system_recovery, **kwargs):
    # unfix variables
    m.fs.splitter.split_fraction[0, "bypass"].unfix()
    m.fs.splitter.split_fraction[0, "bypass"].setlb(0.001)
    m.fs.splitter.split_fraction[0, "bypass"].setub(0.99)

    m.fs.NF.area.unfix()
    m.fs.NF.area.setlb(0.1)
    m.fs.NF.area.setub(1000)

    m.fs.max_conc_factor_target = Param(initialize=3.5, mutable=True)
    m.fs.eq_max_conc_NF = Constraint(
        expr=m.fs.NF.feed_side.properties_out[0].mass_frac_phase_comp["Liq", "Ca"]
        <= m.fs.max_conc_factor_target
        * m.fs.feed.properties[0].mass_frac_phase_comp["Liq", "Ca"]
    )


def set_up_optimization(m, system_recovery=0.50, **kwargs):
    set_optimization_components(m, system_recovery, **kwargs)
    calculate_scaling_factors(m)
    check_dof(m, 2)


def optimize(m, check_termination=True):
    return solve_block(m, tee=True, fail_flag=check_termination)


def optimize_flowsheet(system_recovery=0.50, **kwargs):
    m = solve_flowsheet(**kwargs)
    set_up_optimization(m, system_recovery=system_recovery, **kwargs)
    optimize(m)

    print("===================================" "\n       Optimization            ")
    report(m, **kwargs)

    return m


if __name__ == "__main__":
    m = solve_flowsheet(True)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
import watertap.property_models.water_prop_pack as props
from idaes.models.properties.tests.test_harness import (
    PropertyTestHarness as PropertyTestHarness_idaes,
)
from watertap.property_models.tests.property_test_harness import (
    PropertyTestHarness,
    PropertyRegressionTest,
)

# -----------------------------------------------------------------------------
class TestSeawaterProperty_idaes(PropertyTestHarness_idaes):
    def configure(self):
        self.prop_pack = props.WaterParameterBlock
        self.param_args = {}
        self.prop_args = {}
        self.has_density_terms = False


class TestSeawaterProperty(PropertyTestHarness):
    def configure(self):
        self.prop_pack = props.WaterParameterBlock
        self.param_args = {}
        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1,
        }
        self.stateblock_statistics = {
            "number_variables": 20,
            "number_total_constraints": 16,
            "number_unused_variables": 0,
            "default_degrees_of_freedom": 4,
        }  # 4 state vars
        self.default_solution = {
            ("dens_mass_phase", "Liq"): 996.9,
            ("dens_mass_phase", "Vap"): 0.7363,
            ("flow_vol_phase", "Liq"): 5.016e-4,
            ("flow_vol_phase", "Vap"): 0.6790,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 27.75,
            ("flow_mol_phase_comp", ("Vap", "H2O")): 27.75,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.5,
            ("mole_frac_phase_comp", ("Vap", "H2O")): 0.5,
            ("enth_mass_phase", "Liq"): 1.049e5,
            ("enth_mass_phase", "Vap"): 2.547e6,
            ("dh_vap_mass", None): 2.442e6,
            ("cp_mass_phase", "Liq"): 4.187e3,
            ("cp_mass_phase", "Vap"): 1.865e3,
        }


@pytest.mark.component
class TestSeawaterPropertySolution_1(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.WaterParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1e8,
        }
        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1e-8,
            ("temperature", None): 273.15 + 50,
            ("pressure", None): 2e5,
        }
        self.regression_solution = {
            ("dens_mass_phase", "Liq"): 988.05,
            ("dens_mass_phase", "Vap"): 1.341,
            ("flow_vol_phase", "Liq"): 1.012e-3,
            ("flow_vol_phase", "Vap"): 7.457e-9,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 55.51,
            ("flow_mol_phase_comp", ("Vap", "H2O")): 5.551e-7,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 1,
            ("mole_frac_phase_comp", ("Vap", "H2O")): 1e-8,
            ("enth_mass_phase", "Liq"): 2.094e5,
            ("enth_mass_phase", "Vap"): 2.592e6,
            ("dh_vap_mass", None): 2.382e6,
            ("cp_mass_phase", "Liq"): 4.181e3,
            ("cp_mass_phase", "Vap"): 1.871e3,
        }


@pytest.mark.component
class TestSeawaterPropertySolution_2(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.WaterParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e8,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1,
        }
        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e-8,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1,
            ("temperature", None): 273.15 + 100,
            ("pressure", None): 1e5,
        }
        self.regression_solution = {
            ("dens_mass_phase", "Liq"): 958.3,
            ("dens_mass_phase", "Vap"): 0.5807,
            ("flow_vol_phase", "Liq"): 1.044e-11,
            ("flow_vol_phase", "Vap"): 1.722,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 5.551e-7,
            ("flow_mol_phase_comp", ("Vap", "H2O")): 55.51,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 1e-8,
            ("mole_frac_phase_comp", ("Vap", "H2O")): 1,
            ("enth_mass_phase", "Liq"): 4.190e5,
            ("enth_mass_phase", "Vap"): 2.676e6,
            ("dh_vap_mass", None): 2.257e6,
            ("cp_mass_phase", "Liq"): 4.215e3,
            ("cp_mass_phase", "Vap"): 1.890e3,
        }



import copy
import numpy as np


class _ParameterSweepParallelUtils:
    @classmethod
    def remove_unpicklable_state(cls, parameter_sweep_instance):
        """
        Remove and return any state from the ParameterSweep object that cannot be
        pickled, to make the instance picklable. Needed in order to use the
        ConcurrentFuturesParallelManager.
        """
        saved_state = {
            "parallel_manager": parameter_sweep_instance.parallel_manager,
            "writer": parameter_sweep_instance.writer,
        }

        parameter_sweep_instance.parallel_manager = None
        parameter_sweep_instance.writer = None
        return saved_state

    @classmethod
    def restore_unpicklable_state(cls, parameter_sweep_instance, state):
        """
        Restore a collection of saved state that was removed in order to pickle
        the ParameterSweep object.
        """
        parameter_sweep_instance.parallel_manager = state.get("parallel_manager", None)
        parameter_sweep_instance.writer = state.get("writer", None)

    """
    Combine all of the results retrieved from calling gather().
    - all_results is a list of Result objects, each representing the 
    parameters and results from running the optimization on one process. 
    """

    def _combine_gather_results(self, all_results):
        if len(all_results) == 0:
            return None

        # create the output skeleton based on the first set of results
        # we assume the results are in dict format
        initial_results = all_results[0].results

        combined_results = copy.deepcopy(initial_results)

        # remove any lingering pyomo objects, and convert inner results to numpy arrays
        for key, val in combined_results.items():
            if key != "solve_successful":
                if isinstance(val, dict):
                    for subkey, subval in val.items():
                        if "_pyo_obj" in subval:
                            del subval["_pyo_obj"]

        # for each result, concat the "value" array of results into the
        # gathered results to combine them all

        # get length of data in first result for finding missing keys
        total_chunk_length = len(all_results[0].results["solve_successful"])

        for i, result in enumerate(all_results[1:]):
            results = result.results

            for key, val in results.items():
                if key == "solve_successful":
                    combined_results[key] = np.append(
                        combined_results[key], copy.deepcopy(val)
                    )
                    continue
                if key == "nominal_idx" or key == "differential_idx":
                    combined_results[key] = np.append(
                        combined_results[key], copy.deepcopy(val)
                    )
                    continue
                # print("vall all results!", key, val)
                for subkey, subval in val.items():
                    # lets catch any keys that don' exist in result[0] and
                    # create empty array with expected length, after which we will add
                    # additional values, or add nan's instead
                    if subkey not in combined_results[key]:
                        # create empty array, as none of results so far had this key\

                        combined_results[key][subkey] = {}
                        for sub_subkey, value in subval.items():
                            if sub_subkey == "value":
                                combined_results[key][subkey]["value"] = (
                                    np.zeros(total_chunk_length) * np.nan
                                )
                            else:
                                combined_results[key][subkey][sub_subkey] = value
                    combined_results[key][subkey]["value"] = np.append(
                        combined_results[key][subkey]["value"],
                        copy.deepcopy(
                            subval["value"],
                        ),
                    )
                    # keep track of our subchunk_length
                    sub_chunk_length = len(subval["value"])

                # make sure we add any empty value to missing keys

                for subkey in combined_results[key]:
                    if subkey not in val.keys():
                        empty_chunk = np.zeros(sub_chunk_length) * np.nan
                        combined_results[key][subkey]["value"] = np.append(
                            combined_results[key][subkey]["value"], empty_chunk
                        )
            total_chunk_length += sub_chunk_length
        return combined_results

    """
    Build up a list of the outputs for each result of the optimization.
    Returned as a list of lists, where each inner list is the results from
    one process's run.
    """

    def _combine_output_array(self, gathered_results):
        outputs = gathered_results["outputs"]
        if len(outputs) == 0:
            return []

        # assume all output arrays have the same length
        combined_outputs = [
            np.asarray([]) for _ in range(len(list(outputs.values())[0]["value"]))
        ]
        for _, output in outputs.items():
            for i in range(len(output["value"])):
                combined_outputs[i] = np.append(combined_outputs[i], output["value"][i])
        return np.asarray(combined_outputs)

    """
    Build up a list of the sweep_inputs for each result of the optimization.
    Returned as a list of lists, where each inner list is the results from
    one process's run.
    """

    def _combine_input_array(self, gathered_results):
        inputs = gathered_results["sweep_params"]
        if len(inputs) == 0:
            return []

        # assume all output arrays have the same length
        combined_inputs = [
            np.asarray([]) for _ in range(len(list(inputs.values())[0]["value"]))
        ]
        for _, inputv in inputs.items():
            for i in range(len(inputv["value"])):
                combined_inputs[i] = np.append(combined_inputs[i], inputv["value"][i])
        return np.asarray(combined_inputs)

    """
    Use the embedded ParallelManager to fan out and then back in the results.
    Args:
    - build_model: a function for building the flowsheet model
    - build_model_kwargs: any keyword args necessary for the build_model function
    - build_sweep_params: a function for building the sweep parameters
    - build_sweep_params_kwargs: any keyword args necessary for the build_sweep_params
    function
    - build_outputs: a function for building the outputs dictionary
    - all_parameter_combinations: a list where each element represents the parameters
    for a single local run
    Returns:
    - a list of LocalResults representing the results of the simulation runs 
    """

    def run_scatter_gather(self, all_parameter_combinations, class_reference):
        # save a reference to the parallel manager since it will be removed
        # along with the other unpicklable state
        parallel_manager = self.parallel_manager
        saved_state = class_reference.remove_unpicklable_state(self)

        do_build_kwargs = {"param_sweep_instance": self}

        parallel_manager.scatter(
            do_build,
            do_build_kwargs,
            do_execute,
            all_parameter_combinations,
        )

        # gather the results and combine them into the format we want
        all_results = parallel_manager.gather()
        class_reference.restore_unpicklable_state(self, saved_state)

        return all_results


def do_build(
    param_sweep_instance,
):
    """
    Used to pass into the parallel manager to build the parameters necessary
    for the sweep function. Defined at the top level so it's picklable.
    """
    ps_config = param_sweep_instance.config
    model = ps_config.build_model(**ps_config.build_model_kwargs)
    sweep_params = ps_config.build_sweep_params(
        model, **ps_config.build_sweep_params_kwargs
    )
    sweep_params, sampling_type = param_sweep_instance._process_sweep_params(
        sweep_params
    )
    outputs = ps_config.build_outputs(model, **ps_config.build_outputs_kwargs)
    # for when differential parameter tool is used
    if hasattr(param_sweep_instance, "_define_differential_sweep_outputs"):
        param_sweep_instance._define_differential_sweep_outputs(model, sweep_params)

    if outputs is not None:
        param_sweep_instance.assign_variable_names(model, outputs)

    return [param_sweep_instance, model, sweep_params, outputs]


def do_execute(
    local_combo_array,
    param_sweep_instance,
    model,
    sweep_params,
    outputs,
):
    """
    Used to pass into the parallel manager in order to execute the sweep
    for a set of local values. Defined at the top level so it's picklable.
    """

    if param_sweep_instance.config.custom_do_param_sweep is not None:
        return param_sweep_instance.config.custom_do_param_sweep(
            param_sweep_instance, sweep_params, outputs, local_combo_array
        )

    return param_sweep_instance._do_param_sweep(
        sweep_params, outputs, local_combo_array
    )


def return_none(model, outputkeys=None):
    """
    Used so that build_outputs=None is a valid usage of the parameter sweep tool
    without requiring the user to wrap it in a function.
    """
    return None



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest

from pyomo.environ import ConcreteModel, assert_optimal_termination
from pyomo.util.check_units import assert_units_consistent
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.core.util.scaling as iscale

from watertap.unit_models.mvc.components import Evaporator
import watertap.property_models.seawater_prop_pack as props_sw
import watertap.property_models.water_prop_pack as props_w

solver = get_solver()

# -----------------------------------------------------------------------------
@pytest.mark.requires_idaes_solver
@pytest.mark.component
def test_evaporator():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties_feed = props_sw.SeawaterParameterBlock()
    m.fs.properties_vapor = props_w.WaterParameterBlock()
    m.fs.evaporator = Evaporator(
        property_package_feed=m.fs.properties_feed,
        property_package_vapor=m.fs.properties_vapor,
    )

    # scaling
    m.fs.properties_feed.set_default_scaling(
        "flow_mass_phase_comp", 1, index=("Liq", "H2O")
    )
    m.fs.properties_feed.set_default_scaling(
        "flow_mass_phase_comp", 1e2, index=("Liq", "TDS")
    )
    m.fs.properties_vapor.set_default_scaling(
        "flow_mass_phase_comp", 1, index=("Vap", "H2O")
    )
    m.fs.properties_vapor.set_default_scaling(
        "flow_mass_phase_comp", 1, index=("Liq", "H2O")
    )
    iscale.set_scaling_factor(m.fs.evaporator.area, 1e-3)
    iscale.set_scaling_factor(m.fs.evaporator.U, 1e-3)
    iscale.set_scaling_factor(m.fs.evaporator.delta_temperature_in, 1e-1)
    iscale.set_scaling_factor(m.fs.evaporator.delta_temperature_out, 1e-1)
    iscale.set_scaling_factor(m.fs.evaporator.lmtd, 1e-1)
    iscale.calculate_scaling_factors(m)

    # state variables
    # Feed inlet
    m.fs.evaporator.inlet_feed.flow_mass_phase_comp[0, "Liq", "H2O"].fix(10)
    m.fs.evaporator.inlet_feed.flow_mass_phase_comp[0, "Liq", "TDS"].fix(0.05)
    m.fs.evaporator.inlet_feed.temperature[0].fix(273.15 + 50.52)  # K
    m.fs.evaporator.inlet_feed.pressure[0].fix(1e5)  # Pa

    # Evaporator specifications
    m.fs.evaporator.outlet_brine.temperature[0].fix(273.15 + 60)
    m.fs.evaporator.U.fix(1e3)  # W/K-m^2
    m.fs.evaporator.area.fix(100)  # m^2
    m.fs.evaporator.delta_temperature_out.fix(5)
    m.fs.evaporator.delta_temperature_in.fix(30)

    # check build
    assert_units_consistent(m)
    assert degrees_of_freedom(m) == 0

    # initialize
    m.fs.evaporator.initialize_build()

    # solve
    solver = get_solver()
    results = solver.solve(m, tee=False)
    assert_optimal_termination(results)

    # check values, TODO: make a report for the evaporator
    vapor_blk = m.fs.evaporator.properties_vapor[0]
    assert vapor_blk.flow_mass_phase_comp["Vap", "H2O"].value == pytest.approx(
        0.4183, rel=1e-3
    )
    assert m.fs.evaporator.lmtd.value == pytest.approx(13.79, rel=1e-3)
    assert m.fs.evaporator.heat_transfer.value == pytest.approx(1.379e6, rel=1e-3)
